"""This module implements the Scraper component which parses responses and
extracts information from them"""
import asyncio
from typing import Any, AsyncGenerator, Set, Union, Optional

import aioscrapy
from aioscrapy import signals, Spider
from aioscrapy.exceptions import CloseSpider, DropItem, IgnoreRequest
from aioscrapy.http import PlaywrightResponse
from aioscrapy.http import Request, Response
from aioscrapy.logformatter import LogFormatter
from aioscrapy.middleware import ItemPipelineManager, SpiderMiddlewareManager
from aioscrapy.signalmanager import SignalManager
from aioscrapy.utils.log import logger
from aioscrapy.utils.misc import load_object
from aioscrapy.utils.tools import call_helper, create_task


class Slot:
    """Scraper slot (one per running spider)"""

    MIN_RESPONSE_SIZE = 1024

    def __init__(self, max_active_size: int = 5000000):
        self.max_active_size = max_active_size
        self.active: Set[Request] = set()
        self.active_size: int = 0
        self.itemproc_size: int = 0

    def add_response_request(self, result: Union[Response, BaseException], request: Request) -> None:
        self.active.add(request)
        if isinstance(result, Response):
            self.active_size += max(len(result.body), self.MIN_RESPONSE_SIZE)
        else:
            self.active_size += self.MIN_RESPONSE_SIZE

    def finish_response(self, request: Request, result: Union[Response, BaseException]) -> None:
        self.active.remove(request)
        if isinstance(result, Response):
            self.active_size -= max(len(result.body), self.MIN_RESPONSE_SIZE)
            result._cached_selector = None
        else:
            self.active_size -= self.MIN_RESPONSE_SIZE

    def is_idle(self) -> bool:
        return not self.active

    def needs_backout(self) -> bool:
        return self.active_size > self.max_active_size


class Scraper:

    def __init__(
            self,
            crawler: "aioscrapy.Crawler",
            slot: Slot,
            spidermw: SpiderMiddlewareManager,
            itemproc: ItemPipelineManager,
    ):
        self.crawler = crawler
        self.spider: Spider = crawler.spider
        self.signals: SignalManager = self.crawler.signals
        self.logformatter: LogFormatter = self.crawler.logformatter

        self.slot = slot
        self.spidermw = spidermw
        self.itemproc = itemproc

        self.finish: bool = False
        self.concurrent_parser = asyncio.Semaphore(crawler.settings.getint('CONCURRENT_PARSER', 1))

    @classmethod
    async def from_crawler(cls, crawler: "aioscrapy.Crawler") -> "Scraper":
        instance: "Scraper" = cls(
            crawler,
            Slot(crawler.settings.getint('SCRAPER_SLOT_MAX_ACTIVE_SIZE')),
            await call_helper(SpiderMiddlewareManager.from_crawler, crawler),
            await call_helper(load_object(crawler.settings['ITEM_PROCESSOR']).from_crawler, crawler)
        )
        await instance.itemproc.open_spider(crawler.spider)
        return instance

    async def close(self) -> None:
        """Close a spider being scraped and release its resources"""
        await self.itemproc.close_spider(self.spider)
        self.finish = True

    def is_idle(self) -> bool:
        """Return True if there isn't any more spiders to process"""
        return self.slot.is_idle()

    def needs_backout(self) -> bool:
        return self.slot.needs_backout()

    async def enqueue_scrape(self, result: Union[Response, BaseException], request: Request) -> None:
        # Cache the results in the slot
        self.slot.add_response_request(result, request)
        await self._scrape(result, request)

    async def _scrape(self, result: Union[Response, BaseException], request: Request) -> None:
        """Handle the downloaded response or failure through the spider callback/errback"""
        async with self.concurrent_parser:
            try:
                if not isinstance(result, (Response, BaseException)):
                    raise TypeError(f"Incorrect type: expected Response or Failure, got {type(result)}: {result!r}")
                try:
                    output = await self._scrape2(result, request)  # returns spider's processed output
                except BaseException as e:
                    await self.handle_spider_error(e, request, result)
                else:
                    await self.handle_spider_output(output, request, result)
            except BaseException as e:
                await self.handle_spider_error(e, request, result)
            finally:
                # 控制指纹是否移除
                self.spider.dupefilter and \
                    not request.dont_filter and \
                    await self.spider.dupefilter.done(request, done_type="parse_ok" if getattr(request, "parse_ok", False) else "parse_err")

                if isinstance(result, PlaywrightResponse):
                    await result.release()

                # Delete the cache result from the slot
                self.slot.finish_response(request, result)

    async def _scrape2(self, result: Union[Response, BaseException], request: Request) -> Optional[AsyncGenerator]:
        """Handle the different cases of request's result been a Response or a Exception"""

        if isinstance(result, Response):
            # Throw the response to the middleware of the spider,
            # and the processing results will be processed to the self.call_spider
            return await self.spidermw.scrape_response(self.call_spider, result, request, self.spider)
        else:
            try:
                # Processing Exception of download and download's middleware
                return await self.call_spider(result, request)
            except BaseException as e:
                await self._log_download_errors(e, result, request)

    async def call_spider(self, result: Union[Response, BaseException], request: Request) -> Optional[AsyncGenerator]:
        if isinstance(result, Response):
            # throws Response to Spider's parse
            callback = request.callback or self.spider._parse
            return await call_helper(callback, result, **result.request.cb_kwargs)
        else:
            if request.errback is None:
                raise result
            # throws Exception of download and download's middleware to Spider's errback
            return await call_helper(request.errback, result)

    async def handle_spider_error(self, exc: BaseException, request: Request, response: Response) -> None:
        if isinstance(exc, CloseSpider):
            create_task(self.crawler.engine.close_spider(self.spider, exc.reason or 'cancelled'))
            return
        logger.exception(self.logformatter.spider_error(exc, request, response, self.spider))
        await self.signals.send_catch_log(
            signal=signals.spider_error,
            failure=exc, response=response,
            spider=self.spider
        )
        self.crawler.stats.inc_value("spider_exceptions/%s" % exc.__class__.__name__, spider=self.spider)
        self.crawler.stats.inc_value("spider_exceptions", spider=self.spider)

    async def handle_spider_output(self, result: AsyncGenerator, request: Request, response: Response) -> None:
        """Iter each Request/Item (given in the output parameter) returned from the given spider"""
        if not result:
            return

        parse_ok = True
        while True:
            try:
                output = await result.__anext__()
            except StopAsyncIteration:
                break
            except Exception as e:
                parse_ok = False
                await self.handle_spider_error(e, request, response)
            else:
                await self._process_spidermw_output(output, request, response)

        self.spider.dupefilter and \
            not request.dont_filter and \
            setattr(request, "parse_ok", parse_ok)

    async def _process_spidermw_output(self, output: Any, request: Request, response: Response) -> None:
        """Process each Request/Item (given in the output parameter) returned from the given spider"""

        if isinstance(output, Request):
            await self.crawler.engine.crawl(request=output)
        elif isinstance(output, dict):
            self.slot.itemproc_size += 1
            item = await self.itemproc.process_item(output, self.spider)
            process_item_method = getattr(self.spider, 'process_item', None)
            if process_item_method:
                await call_helper(process_item_method, item)
            await self._itemproc_finished(output, item, response)
        elif output is None:
            pass
        else:
            typename = type(output).__name__
            logger.error(
                'Spider must return request, item, or None, got %(typename)r in %(request)s' % {'request': request,
                                                                                                'typename': typename},
            )

    async def _log_download_errors(
            self,
            spider_exception: BaseException,
            download_exception: BaseException,
            request: Request
    ) -> None:
        """Process and record errors"""
        if isinstance(download_exception, BaseException) and not isinstance(download_exception, IgnoreRequest):
            logger.exception(self.logformatter.download_error(download_exception, request, self.spider))

        if spider_exception is not download_exception:
            raise spider_exception

    async def _itemproc_finished(self, output: Any, item: Any, response: Response) -> None:
        """ItemProcessor finished for the given ``item`` and returned ``output``"""
        self.slot.itemproc_size -= 1
        if isinstance(output, BaseException):
            if isinstance(output, DropItem):
                logger.log(**self.logformatter.dropped(item, output, response, self.spider))
                return await self.signals.send_catch_log_deferred(
                    signal=signals.item_dropped, item=item, response=response,
                    spider=self.spider, exception=output)
            else:
                logger.exception(self.logformatter.item_error(item, output, response, self.spider))
                return await self.signals.send_catch_log_deferred(
                    signal=signals.item_error, item=item, response=response,
                    spider=self.spider, failure=output)
        else:
            logger.log(**self.logformatter.scraped(output, response, self.spider))
            return await self.signals.send_catch_log_deferred(
                signal=signals.item_scraped, item=output, response=response,
                spider=self.spider)



from typing import Literal

from aioscrapy import Request
from aioscrapy.db import db_manager
from aioscrapy.dupefilters import DupeFilterBase


class RedisRFPDupeFilter(DupeFilterBase):
    """Request Fingerprint duplicates filter built with Set of Redis"""

    def __init__(
            self,
            server: "redis.asyncio.Redis",
            key: str,
            debug: bool = False,
            keep_on_close: bool = True,
            info: bool = False,
    ):
        self.server = server
        self.key = key
        self.debug = debug
        self.keep_on_close = keep_on_close
        self.logdupes: bool = True
        self.info: bool = info

    @classmethod
    def from_crawler(cls, crawler: "aioscrapy.crawler.Crawler"):
        server = db_manager.redis.queue
        dupefilter_key = crawler.settings.get("SCHEDULER_DUPEFILTER_KEY", '%(spider)s:dupefilter')
        keep_on_close = crawler.settings.getbool("KEEP_DUPEFILTER_DATA_ON_CLOSE", True)
        key = dupefilter_key % {'spider': crawler.spider.name}
        debug = crawler.settings.getbool('DUPEFILTER_DEBUG', False)
        info = crawler.settings.getbool('DUPEFILTER_INFO', False)
        instance = cls(server, key=key, debug=debug, keep_on_close=keep_on_close, info=info)
        return instance

    async def request_seen(self, request: Request):
        return await self.server.sadd(self.key, request.fingerprint) == 0

    async def close(self, reason=''):
        if not self.keep_on_close:
            await self.clear()

    async def clear(self):
        await self.server.delete(self.key)


class HashMap(object):
    def __init__(self, m, seed):
        self.m = m
        self.seed = seed

    def hash(self, value):
        """
        Hash Algorithm
        :param value: Value
        :return: Hash Value
        """
        ret = 0
        for i in range(len(value)):
            ret += self.seed * ret + ord(value[i])
        return (self.m - 1) & ret


class BloomFilter(object):
    def __init__(self, server, key, bit=30, hash_number=6):
        """
        Initialize BloomFilter
        :param server: Redis Server
        :param key: BloomFilter Key
        :param bit: m = 2 ^ bit
        :param hash_number: the number of hash function
        """
        # default to 1 << 30 = 10,7374,1824 = 2^30 = 128MB, max filter 2^30/hash_number = 1,7895,6970 fingerprints
        self.m = 1 << bit
        self.seeds = range(hash_number)
        self.server = server
        self.key = key
        self.maps = [HashMap(self.m, seed) for seed in self.seeds]

    async def exists(self, value):
        if not value:
            return False
        async with self.server.pipeline(transaction=True) as pipe:
            for f in self.maps:
                offset = f.hash(value)
                pipe.getbit(self.key, offset)
            result = await pipe.execute()
        return all(result)

    async def insert(self, value):
        """
        add value to bloom
        :param value:
        :return:
        """
        async with self.server.pipeline(transaction=True) as pipe:
            for f in self.maps:
                offset = f.hash(value)
                pipe.setbit(self.key, offset, 1)
            await pipe.execute()


class RedisBloomDupeFilter(RedisRFPDupeFilter):
    """Bloom filter built with the bitis bitmap of redis"""

    def __init__(self, server, key, debug, bit, hash_number, keep_on_close, info):
        super().__init__(server, key, debug, keep_on_close, info)
        self.bit = bit
        self.hash_number = hash_number
        self.bf = BloomFilter(server, self.key, bit, hash_number)

    @classmethod
    async def from_crawler(cls, crawler: "aioscrapy.crawler.Crawler"):
        server = db_manager.redis.queue
        dupefilter_key = crawler.settings.get("SCHEDULER_DUPEFILTER_KEY", '%(spider)s:bloomfilter')
        keep_on_close = crawler.settings.getbool("KEEP_DUPEFILTER_DATA_ON_CLOSE", True)
        key = dupefilter_key % {'spider': crawler.spider.name}
        debug = crawler.settings.getbool('DUPEFILTER_DEBUG', False)
        info = crawler.settings.getbool('DUPEFILTER_INFO', False)
        bit = crawler.settings.getint('BLOOMFILTER_BIT', 30)
        hash_number = crawler.settings.getint('BLOOMFILTER_HASH_NUMBER', 6)
        return cls(server, key=key, debug=debug, bit=bit, hash_number=hash_number, keep_on_close=keep_on_close, info=info)

    async def request_seen(self, request: Request) -> bool:
        fp = await self.bf.exists(request.fingerprint)
        if fp:
            return True
        await self.bf.insert(request.fingerprint)
        return False


class ExRedisBloomDupeFilter(RedisBloomDupeFilter):

    def __init__(self, server, key, key_set, ttl, debug, bit, hash_number, keep_on_close, info):
        super().__init__(server, key, debug, bit, hash_number, keep_on_close, info)
        self.key_set = key_set
        self.ttl = ttl

    @classmethod
    async def from_crawler(cls, crawler: "aioscrapy.crawler.Crawler"):
        server = db_manager.redis.queue
        dupefilter_key = crawler.settings.get("SCHEDULER_DUPEFILTER_KEY", '%(spider)s:bloomfilter')
        keep_on_close = crawler.settings.getbool("KEEP_DUPEFILTER_DATA_ON_CLOSE", True)
        key = dupefilter_key % {'spider': crawler.spider.name}
        debug = crawler.settings.getbool('DUPEFILTER_DEBUG', False)
        info = crawler.settings.getbool('DUPEFILTER_INFO', False)
        bit = crawler.settings.getint('BLOOMFILTER_BIT', 30)
        hash_number = crawler.settings.getint('BLOOMFILTER_HASH_NUMBER', 6)
        ttl = crawler.settings.getint('DUPEFILTER_SET_KEY_TTL', 180)
        return cls(server, key=key, key_set=key + "_set", ttl=ttl, debug=debug, bit=bit, hash_number=hash_number,
                   keep_on_close=keep_on_close, info=info)

    async def request_seen(self, request: Request) -> bool:
        fp = await self.bf.exists(request.fingerprint)
        if fp:
            return True
        async with self.server.pipeline() as pipe:
            pipe.sadd(self.key_set, request.fingerprint)
            pipe.expire(self.key_set, self.ttl)
            ret, _ = await pipe.execute()
        return ret == 0

    async def done(
            self,
            request: Request,
            done_type: Literal["request_ok", "request_err", "parse_ok", "parse_err"]
    ):
        if done_type == "request_ok" or done_type == "request_err":
            await self.server.srem(self.key_set, request.fingerprint)
        elif done_type == "parse_ok":
            await self.bf.insert(request.fingerprint)

    async def close(self, reason=''):
        if not self.keep_on_close:
            await self.clear()
        await self.server.delete(self.key_set)


class ExRedisRFPDupeFilter(RedisRFPDupeFilter):

    async def done(
            self,
            request: Request,
            done_type: Literal["request_ok", "request_err", "parse_ok", "parse_err"]
    ):
        # 当请求失败或解析失败的时候 从Redis的Set中移除指纹
        if done_type == "request_err" or done_type == "parse_err":
            await self.server.srem(self.key, request.fingerprint)


RFPDupeFilter = RedisRFPDupeFilter
ExRFPDupeFilter = ExRedisRFPDupeFilter
BloomDupeFilter = RedisBloomDupeFilter
ExBloomDupeFilter = ExRedisBloomDupeFilter
BloomSetDupeFilter = ExRedisBloomDupeFilter




import asyncio
import sys
import warnings
from typing import Type

from loguru import logger as _logger

from aioscrapy.settings import Settings

for _handler in _logger._core.handlers.values():
    if _handler._name == '<stderr>':
        _logger.remove(_handler._id)


def configure_logging(spider: Type["Spider"], settings: Settings):
    formatter = settings.get('LOG_FORMAT')
    level = settings.get('LOG_LEVEL', 'INFO')
    enqueue = settings.get('ENQUEUE', True)
    if settings.get('LOG_STDOUT', True):
        _logger.add(
            sys.stderr, format=formatter, level=level, enqueue=enqueue,
            filter=lambda record: record["extra"].get("spidername") == spider.name,
        )

    if filename := settings.get('LOG_FILE'):
        rotation = settings.get('LOG_ROTATION', '20MB')
        retention = settings.get('LOG_RETENTION', 10)
        encoding = settings.get('LOG_ENCODING', 'utf-8')
        _logger.add(
            sink=filename, format=formatter, encoding=encoding, level=level,
            enqueue=enqueue, rotation=rotation, retention=retention,
            filter=lambda record: record["extra"].get("spidername") == spider.name,
        )


class AioScrapyLogger:
    __slots__ = (
        'catch', 'complete', 'critical', 'debug', 'error', 'exception',
        'info', 'log', 'patch', 'success', 'trace', 'warning'
    )

    def __getattr__(self, method):
        try:
            spider_name = asyncio.current_task().get_name()
            return getattr(_logger.bind(spidername=spider_name), method)
        except Exception as e:
            warnings.warn(f'Error on get logger: {e}')
            return getattr(_logger, method)


logger = AioScrapyLogger()



"""
This module provides some useful functions for working with
aioscrapy.http.Request objects
"""

from typing import Optional
from urllib.parse import urlunparse

from w3lib.http import headers_dict_to_raw

from aioscrapy import Spider, Request
from aioscrapy.utils.httpobj import urlparse_cached
from aioscrapy.utils.misc import load_object
from aioscrapy.utils.python import to_bytes, to_unicode


def request_httprepr(request: Request) -> bytes:
    """Return the raw HTTP representation (as bytes) of the given request.
    This is provided only for reference since it's not the actual stream of
    bytes that will be send when performing the request (that's controlled
    by Twisted).
    """
    parsed = urlparse_cached(request)
    path = urlunparse(('', '', parsed.path or '/', parsed.params, parsed.query, ''))
    s = to_bytes(request.method) + b" " + to_bytes(path) + b" HTTP/1.1\r\n"
    s += b"Host: " + to_bytes(parsed.hostname or b'') + b"\r\n"
    if request.headers:
        s += headers_dict_to_raw({to_bytes(k): to_bytes(v) for k, v in request.headers.items()}) + b"\r\n"
    s += b"\r\n"
    s += to_bytes(request.body)
    return s


def referer_str(request: Request) -> Optional[str]:
    """ Return Referer HTTP header suitable for logging. """
    referrer = request.headers.get('Referer')
    if referrer is None:
        return referrer
    return to_unicode(referrer, errors='replace')


async def request_from_dict(d: dict, *, spider: Optional[Spider] = None) -> Request:
    """Create a :class:`~scrapy.Request` object from a dict.

    If a spider is given, it will try to resolve the callbacks looking at the
    spider for methods with the same name.
    """
    d = await spider.request_from_dict(d) or d
    if isinstance(d, Request):
        return d

    request_cls = load_object(d["_class"]) if "_class" in d else Request
    kwargs = {key: value for key, value in d.items() if key in request_cls.attributes}
    if d.get("callback") and spider:
        kwargs["callback"] = _get_method(spider, d["callback"])
    if d.get("errback") and spider:
        kwargs["errback"] = _get_method(spider, d["errback"])
    return request_cls(**kwargs)


def _get_method(obj, name):
    """Helper function for request_from_dict"""
    name = str(name)
    try:
        return getattr(obj, name)
    except AttributeError:
        raise ValueError(f"Method {name!r} not found in: {obj}")



import asyncio

import pyhttpx
from pyhttpx.exception import BaseExpetion as PyHttpxError

from aioscrapy import Request
from aioscrapy.core.downloader.handlers import BaseDownloadHandler
from aioscrapy.exceptions import DownloadError
from aioscrapy.http import HtmlResponse
from aioscrapy.settings import Settings
from aioscrapy.utils.log import logger


class PyhttpxDownloadHandler(BaseDownloadHandler):

    def __init__(self, settings):
        self.settings: Settings = settings
        self.pyhttpx_client_args: dict = self.settings.get('PYHTTPX_CLIENT_ARGS', {})
        self.verify_ssl = self.settings.get("VERIFY_SSL", True)
        self.loop = asyncio.get_running_loop()

    @classmethod
    def from_settings(cls, settings: Settings):
        return cls(settings)

    async def download_request(self, request: Request, _) -> HtmlResponse:
        try:
            return await self._download_request(request)
        except PyHttpxError as e:
            raise DownloadError(real_error=e) from e

    async def _download_request(self, request: Request) -> HtmlResponse:
        kwargs = {
            'timeout': self.settings.get('DOWNLOAD_TIMEOUT'),
            'cookies': dict(request.cookies),
            'verify': self.verify_ssl,
            'allow_redirects': self.settings.getbool('REDIRECT_ENABLED', True) if request.meta.get(
                'dont_redirect') is None else request.meta.get('dont_redirect')
        }
        post_data = request.body or None
        if isinstance(post_data, dict):
            kwargs['json'] = post_data
        else:
            kwargs['data'] = post_data

        headers = request.headers or self.settings.get('DEFAULT_REQUEST_HEADERS')
        kwargs['headers'] = headers

        proxy = request.meta.get("proxy")
        if proxy:
            kwargs["proxies"] = {'https': proxy}
            logger.debug(f"use proxy {proxy}: {request.url}")

        session_args = self.pyhttpx_client_args.copy()
        session_args.setdefault('http2', True)
        with pyhttpx.HttpSession(**session_args) as session:
            response = await asyncio.to_thread(session.request, request.method, request.url, **kwargs)
            return HtmlResponse(
                request.url,
                status=response.status_code,
                headers=response.headers,
                body=response.content,
                cookies=dict(response.cookies),
                encoding=response.encoding
            )

    async def close(self):
        pass



import inspect

from aioscrapy.spiders import Spider


def iter_spider_classes(module):
    """Return an iterator over all spider classes defined in the given module
    that can be instantiated (i.e. which have name)
    """
    # this needs to be imported here until get rid of the spider manager
    # singleton in aioscrapy.spider.spiders

    for obj in vars(module).values():
        if (
                inspect.isclass(obj)
                and issubclass(obj, Spider)
                and obj.__module__ == module.__name__
                and getattr(obj, 'name', None)
        ):
            yield obj



"""
Base class for Scrapy spiders

See documentation in docs/topics/spiders.rst
"""
import time
from typing import Optional, Union

from aioscrapy import signals
from aioscrapy.exceptions import DontCloseSpider
from aioscrapy.http.request import Request
from aioscrapy.http.response import Response
from aioscrapy.statscollectors import StatsCollector
from aioscrapy.utils.tools import call_helper
from aioscrapy.utils.url import url_is_from_spider


class Spider(object):
    """Base class for scrapy spiders. All spiders must inherit from this
    class.
    """

    name: Optional[str] = None
    proxy: Optional["aioscrapy.proxy.AbsProxy"] = None
    dupefilter: Optional["aioscrapy.dupefilters.DupeFilterBase"] = None
    custom_settings: Optional[dict] = None
    stats: Optional[StatsCollector] = None

    pause: bool = False
    _pause_time: Optional[Union[int, float]] = None

    def __init__(self, name=None, **kwargs):
        if name is not None:
            self.name = name
        elif not getattr(self, 'name', None):
            raise ValueError(f"{type(self).__name__} must have a name")
        self.__dict__.update(kwargs)
        if not hasattr(self, 'start_urls'):
            self.start_urls = []

    @property
    def pause_time(self) -> int:
        if self._pause_time is None:
            self._pause_time = 600 + int(time.time())
        return self._pause_time

    @pause_time.setter
    def pause_time(self, value: Union[int, float]):
        self.pause = True
        if value is None:
            self._pause_time = float('inf')
        elif value < time.time():
            self._pause_time = value + int(time.time())
        else:
            self._pause_time = value

    @classmethod
    async def from_crawler(cls, crawler, *args, **kwargs):
        spider = cls(*args, **kwargs)
        spider._set_crawler(crawler)
        return spider

    def _set_crawler(self, crawler):
        self.crawler = crawler
        self.settings = crawler.settings
        self.close_on_idle = self.settings.get("CLOSE_SPIDER_ON_IDLE", True)
        crawler.signals.connect(self.close, signals.spider_closed)
        crawler.signals.connect(self.spider_idle, signal=signals.spider_idle)

    async def start_requests(self):
        if not self.start_urls and hasattr(self, 'start_url'):
            raise AttributeError(
                "Crawling could not start: 'start_urls' not found "
                "or empty (but found 'start_url' attribute instead, "
                "did you miss an 's'?)")

        for url in self.start_urls:
            yield Request(url)

    async def request_from_dict(self, d: dict):
        """继承成后重写改方法，将队列中的json根据情况构建成Request对象"""
        pass

    async def _parse(self, response: Response, **kwargs):
        return await call_helper(self.parse, response)

    async def parse(self, response: Response):
        raise NotImplementedError(f'{self.__class__.__name__}.parse callback is not defined')

    @classmethod
    def update_settings(cls, settings):
        settings.setdict(cls.custom_settings or {}, priority='spider')

    @classmethod
    def handles_request(cls, request):
        return url_is_from_spider(request.url, cls)

    @staticmethod
    def close(spider, reason):
        closed = getattr(spider, 'closed', None)
        if callable(closed):
            return closed(reason)

    def __str__(self):
        return f"<{type(self).__name__} {self.name!r} at 0x{id(self):0x}>"

    __repr__ = __str__

    @classmethod
    def start(cls, setting_path=None, use_windows_selector_eventLoop: bool = False):
        from aioscrapy.crawler import CrawlerProcess
        from aioscrapy.utils.project import get_project_settings

        settings = get_project_settings()
        if setting_path is not None:
            settings.setmodule(setting_path)
        cp = CrawlerProcess(settings)
        cp.crawl(cls)
        cp.start(use_windows_selector_eventLoop)

    def spider_idle(self):
        if not self.close_on_idle:
            raise DontCloseSpider



import asyncio
import random
from abc import abstractmethod
from collections import deque
from datetime import datetime
from time import time
from typing import Optional, Set, Deque, Tuple, Callable, TypeVar

from aioscrapy import signals, Request, Spider
from aioscrapy.core.downloader.handlers import DownloadHandlerManager
from aioscrapy.dupefilters import DupeFilterBase
from aioscrapy.http import Response
from aioscrapy.middleware import DownloaderMiddlewareManager
from aioscrapy.proxy import AbsProxy
from aioscrapy.settings import Settings
from aioscrapy.signalmanager import SignalManager
from aioscrapy.utils.httpobj import urlparse_cached
from aioscrapy.utils.log import logger
from aioscrapy.utils.misc import load_instance
from aioscrapy.utils.tools import call_helper, create_task


class BaseDownloaderMeta(type):

    def __instancecheck__(cls, instance):
        return cls.__subclasscheck__(type(instance))

    def __subclasscheck__(cls, subclass):
        return (
                hasattr(subclass, "fetch") and callable(subclass.fetch)
                and hasattr(subclass, "needs_backout") and callable(subclass.needs_backout)
        )


class BaseDownloader(metaclass=BaseDownloaderMeta):

    @classmethod
    async def from_crawler(cls, crawler) -> "BaseDownloader":
        return cls()

    async def close(self) -> None:
        pass

    @abstractmethod
    async def fetch(self, request: Request) -> None:
        raise NotImplementedError()

    @abstractmethod
    def needs_backout(self) -> bool:
        raise NotImplementedError()


DownloaderTV = TypeVar("DownloaderTV", bound="Downloader")


class Slot:
    """Downloader slot"""

    def __init__(self, concurrency: int, delay: float, randomize_delay: bool) -> None:
        self.concurrency = concurrency
        self.delay = delay
        self.randomize_delay = randomize_delay

        self.active: Set[Request] = set()
        self.transferring: Set[Request] = set()
        self.queue: Deque[Request] = deque()
        self.lastseen: float = 0
        self.delay_lock: bool = False

    def free_transfer_slots(self):
        return self.concurrency - len(self.transferring)

    def download_delay(self) -> float:
        if self.randomize_delay:
            return random.uniform(0.5 * self.delay, 1.5 * self.delay)
        return self.delay

    def __repr__(self):
        cls_name = self.__class__.__name__
        return "%s(concurrency=%r, delay=%0.2f, randomize_delay=%r)" % (
            cls_name, self.concurrency, self.delay, self.randomize_delay)

    def __str__(self):
        return (
                "<downloader.Slot concurrency=%r delay=%0.2f randomize_delay=%r "
                "len(active)=%d len(queue)=%d len(transferring)=%d lastseen=%s>" % (
                    self.concurrency, self.delay, self.randomize_delay,
                    len(self.active), len(self.queue), len(self.transferring),
                    datetime.fromtimestamp(self.lastseen).isoformat()
                )
        )


def _get_concurrency_delay(concurrency: int, spider: Spider, settings: Settings) -> Tuple[int, float]:
    delay = settings.getfloat('DOWNLOAD_DELAY')
    if hasattr(spider, 'download_delay'):
        delay = spider.download_delay

    if hasattr(spider, 'max_concurrent_requests'):
        concurrency = spider.max_concurrent_requests

    return concurrency, delay


class Downloader(BaseDownloader):
    DOWNLOAD_SLOT: str = 'download_slot'

    def __init__(
            self,
            crawler,
            handler: DownloadHandlerManager,
            middleware: DownloaderMiddlewareManager,
            *,
            proxy: Optional[AbsProxy] = None,
            dupefilter: Optional[DupeFilterBase] = None,
    ):
        self.settings: Settings = crawler.settings
        self.signals: SignalManager = crawler.signals
        self.spider: Spider = crawler.spider
        self.spider.proxy = proxy
        self._call_engine: Callable = crawler.engine.handle_downloader_output

        self.middleware = middleware
        self.handler = handler
        self.proxy = proxy
        self.dupefilter = dupefilter

        self.total_concurrency: int = self.settings.getint('CONCURRENT_REQUESTS')
        self.get_requests_count: int = self.settings.getint('GET_REQUESTS_COUNT') or self.total_concurrency
        self.domain_concurrency: int = self.settings.getint('CONCURRENT_REQUESTS_PER_DOMAIN')
        self.ip_concurrency: int = self.settings.getint('CONCURRENT_REQUESTS_PER_IP')
        self.randomize_delay: bool = self.settings.getbool('RANDOMIZE_DOWNLOAD_DELAY')

        self.active: Set[Request] = set()
        self.slots: dict = {}
        self.running: bool = True
        create_task(self._slot_gc(60))

    @classmethod
    async def from_crawler(cls, crawler) -> "Downloader":
        df = crawler.settings.get('DUPEFILTER_CLASS') and await load_instance(crawler.settings['DUPEFILTER_CLASS'],
                                                                              crawler=crawler)
        crawler.spider.dupefilter = df  # 将指纹绑定到Spider 在解析成功的时候 调用DUPEFILTER_CLASS的success方法
        return cls(
            crawler,
            await call_helper(DownloadHandlerManager.from_crawler, crawler),
            await call_helper(DownloaderMiddlewareManager.from_crawler, crawler),
            proxy=crawler.settings.get("PROXY_HANDLER") and await load_instance(crawler.settings["PROXY_HANDLER"],
                                                                                crawler=crawler),
            dupefilter=df
        )

    async def fetch(self, request: Request) -> None:
        self.active.add(request)
        key, slot = self._get_slot(request, self.spider)
        request.meta[self.DOWNLOAD_SLOT] = key

        slot.active.add(request)
        slot.queue.append(request)
        await self._process_queue(slot)

    async def _process_queue(self, slot: Slot) -> None:
        if slot.delay_lock:
            return

        now = time()
        delay = slot.download_delay()
        if delay:
            penalty = delay - now + slot.lastseen
            if penalty > 0:
                slot.delay_lock = True
                await asyncio.sleep(penalty)
                slot.delay_lock = False
                create_task(self._process_queue(slot))
                return

        while slot.queue and slot.free_transfer_slots() > 0:
            request = slot.queue.popleft()
            slot.transferring.add(request)
            create_task(self._download(slot, request))
            if delay:
                break

    async def _download(self, slot: Slot, request: Request) -> None:
        result = None
        try:
            if self.dupefilter and not request.dont_filter and await self.dupefilter.request_seen(request):
                self.dupefilter.log(request, self.spider)
                return
            slot.lastseen = time()
            result = await self.middleware.process_request(self.spider, request)
            if result is None:
                self.proxy and await self.proxy.add_proxy(request)
                result = await self.handler.download_request(request, self.spider)
        except BaseException as exc:
            self.proxy and self.proxy.check(request, exception=exc)
            result = await self.middleware.process_exception(self.spider, request, exc)
        else:
            if isinstance(result, Response):
                try:
                    self.proxy and self.proxy.check(request, response=result)
                    result = await self.middleware.process_response(self.spider, request, result)
                except BaseException as exc:
                    result = exc
        finally:
            slot.transferring.remove(request)
            slot.active.remove(request)
            self.active.remove(request)

            if isinstance(result, Response):
                await self.signals.send_catch_log(signal=signals.response_downloaded,
                                                  response=result,
                                                  request=request,
                                                  spider=self.spider)
            #  控制指纹是否移除
            self.dupefilter and \
                not request.dont_filter and \
                await self.dupefilter.done(request, done_type="request_ok" if isinstance(result, Response) else "request_err")

            await self._call_engine(result, request)
            await self._process_queue(slot)

    async def close(self) -> None:
        self.running = False
        self.dupefilter and await self.dupefilter.close()

    async def _slot_gc(self, age=60):
        while self.running:
            await asyncio.sleep(age)
            for key, slot in list(self.slots.items()):
                logger.debug(slot)
                if not slot.active and slot.lastseen + slot.delay < (time() - age):
                    self.slots.pop(key)

    def needs_backout(self):
        return len(self.active) >= self.total_concurrency

    def _get_slot(self, request, spider):
        key = self._get_slot_key(request, spider)
        if key not in self.slots:
            conc = self.ip_concurrency if self.ip_concurrency else self.domain_concurrency
            conc, delay = _get_concurrency_delay(conc, spider, self.settings)
            self.slots[key] = Slot(conc, delay, self.randomize_delay)
        return key, self.slots[key]

    def _get_slot_key(self, request, spider):
        if self.DOWNLOAD_SLOT in request.meta:
            return request.meta[self.DOWNLOAD_SLOT]

        if self.ip_concurrency:
            return request.meta.get("proxy", '')
        else:
            return urlparse_cached(request).hostname or ''



import pprint
from abc import ABCMeta, abstractmethod
from collections import defaultdict, deque

from aioscrapy.exceptions import NotConfigured
from aioscrapy.utils.log import logger
from aioscrapy.utils.misc import load_instance
from aioscrapy.utils.tools import call_helper


class AbsMiddlewareManager(object, metaclass=ABCMeta):
    """Base class for implementing middleware managers"""

    component_name = 'foo middleware'

    def __init__(self, *middlewares):
        self.middlewares = middlewares
        self.methods = defaultdict(deque)
        for mw in middlewares:
            self._add_middleware(mw)

    @classmethod
    @abstractmethod
    def _get_mwlist_from_settings(cls, settings):
        """get middleware list from settings"""

    @classmethod
    async def from_settings(cls, settings, crawler=None):
        mwlist = cls._get_mwlist_from_settings(settings)
        middlewares = []
        enabled = []
        for clspath in mwlist:
            try:
                middlewares.append(await load_instance(clspath, settings=settings, crawler=crawler))
                enabled.append(clspath)
            except NotConfigured as e:
                if e.args:
                    clsname = clspath.split('.')[-1]
                    logger.warning("Disabled %(clsname)s: %(eargs)s" % {'clsname': clsname, 'eargs': e.args[0]})

        logger.info(f"Enabled {cls.component_name}s:\n{pprint.pformat(enabled)}")
        return cls(*middlewares)

    @classmethod
    async def from_crawler(cls, crawler):
        return await cls.from_settings(crawler.settings, crawler)

    def _add_middleware(self, mw):
        if hasattr(mw, 'open_spider'):
            self.methods['open_spider'].append(mw.open_spider)
        if hasattr(mw, 'close_spider'):
            self.methods['close_spider'].appendleft(mw.close_spider)

    async def _process_parallel(self, methodname, obj, *args):
        return await self.process_parallel(self.methods[methodname], obj, *args)

    async def _process_chain(self, methodname, obj, *args):
        return await self.process_chain(self.methods[methodname], obj, *args)

    async def _process_chain_both(self, cb_methodname, eb_methodname, obj, *args):
        return await self.process_chain_both(self.methods[cb_methodname],
                                             self.methods[eb_methodname], obj, *args)

    async def open_spider(self, spider):
        return await self._process_parallel('open_spider', spider)

    async def close_spider(self, spider):
        return await self._process_parallel('close_spider', spider)

    @staticmethod
    async def process_parallel(callbacks, input_, *a, **kw):
        for callback in callbacks:
            await call_helper(callback, input_, *a, **kw)

    @staticmethod
    async def process_chain(callbacks, input_, *a, **kw):
        for callback in callbacks:
            input_result = await call_helper(callback, input_, *a, **kw)
            if input_result is not None:
                input_ = input_result
        return input_

    @staticmethod
    async def process_chain_both(callbacks, errbacks, input_, *a, **kw):
        for cb, eb in zip(callbacks, errbacks):
            try:
                input_ = await call_helper(cb, input_, *a, **kw)
            except(Exception, BaseException):
                input_ = await call_helper(eb, input_, *a, **kw)
            return input_



# -*- coding: utf-8 -*-

import os
from typing import Dict, Optional, Tuple

try:
    from typing import Literal  # python >= 3.8
except ImportError:  # python <3.8
    from typing_extensions import Literal

from urllib.parse import urlparse, urlunparse

from playwright.async_api import Page, BrowserContext, ViewportSize, ProxySettings
from playwright.async_api import Playwright, Browser
from playwright.async_api import async_playwright


class PlaywrightDriver:
    def __init__(
            self,
            *,
            driver_type: Literal["chromium", "firefox", "webkit"] = "chromium",
            proxy: Optional[str] = None,
            browser_args: Optional[Dict] = None,
            context_args: Optional[Dict] = None,
            window_size: Optional[Tuple[int, int]] = None,
            user_agent: str = None,
            **kwargs
    ):

        self.driver_type = driver_type
        self.proxy = proxy and self.format_context_proxy(proxy)
        self.viewport = window_size and ViewportSize(width=window_size[0], height=window_size[1])
        self.browser_args = browser_args or {}
        self.context_args = context_args or {}
        self.user_agent = user_agent

        self.driver: Optional[Playwright] = None
        self.browser: Optional[Browser] = None
        self.context: Optional[BrowserContext] = None
        self.page: Optional[Page] = None
        self.url = None

    async def setup(self):
        browser_args = self.browser_args.copy()
        context_args = self.context_args.copy()
        if browser_args.get('args') is None:
            browser_args.update({'args': ["--no-sandbox"]})

        if context_args.get("storage_state") is not None:
            storage_state_path = context_args.get("storage_state")
            os.makedirs(os.path.dirname(storage_state_path), exist_ok=True)

        if self.proxy:
            browser_args.update({'proxy': self.proxy})
            context_args.update({'proxy': self.proxy})
        if self.viewport:
            context_args.update({"viewport": self.viewport})
            context_args.update({"screen": self.viewport})
        if self.user_agent:
            context_args.update({'user_agent': self.user_agent})

        self.driver = await async_playwright().start()
        self.browser: Browser = await getattr(self.driver, self.driver_type).launch(**browser_args)
        self.context = await self.browser.new_context(**context_args)
        self.page = await self.context.new_page()

    @staticmethod
    def format_context_proxy(proxy) -> ProxySettings:
        parsed_url = urlparse(proxy)
        return ProxySettings(
            server=urlunparse(parsed_url._replace(netloc=parsed_url.netloc.split('@')[-1])),
            username=parsed_url.username,
            password=parsed_url.password,
        )

    async def quit(self):
        await self.page.close()
        try:
            await self.context.close()
        except:
            pass
        finally:
            await self.browser.close()
            await self.driver.stop()

    async def get_cookies(self):
        return {
            cookie["name"]: cookie["value"]
            for cookie in await self.page.context.cookies()
        }

    async def set_cookies(self, cookies: dict):
        await self.page.context.add_cookies([
            {"name": key, "value": value, "url": self.url or self.page.url} for key, value in cookies.items()
        ])



import asyncio
import csv
from typing import Optional

from aioscrapy.utils.log import logger


class CsvSinkMixin:
    csv_writer = {}

    async def save_item(
            self,
            item: dict,
            *,
            filename: Optional[str] = None,
    ):
        assert filename is not None, "请传入filename参数"
        if '.csv' not in filename:
            filename = filename + '.csv'
        try:
            writer = self._get_writer(filename, item)
            writer.writerow(item.values())
        except Exception as e:
            logger.exception(f'Save csv Error, filename:{filename}, item:{item}, errMsg: {e}')

    def _get_writer(self, filename, item):
        writer, *_ = self.csv_writer.get(filename, (None, None))
        if writer is None:
            file = open(filename, 'w', encoding="UTF8", newline='')
            writer = csv.writer(file)
            writer.writerow(item.keys())
            self.csv_writer[filename] = (writer, file)
        return writer

    def close_csv(self, filename=None):
        *_, file = self.csv_writer.pop(filename, (None, None))
        if file is not None:
            logger.info(f'Closing csv: {filename}')
            file.close()

    def close(self):
        for filename in list(self.csv_writer.keys()):
            self.close_csv(filename)


class CsvPipeline(CsvSinkMixin):
    def __init__(self, settings):
        self.lock = asyncio.Lock()

    @classmethod
    def from_settings(cls, settings):
        return cls(settings)

    async def process_item(self, item, spider):
        execl_kw: Optional[dict] = item.pop('__csv__', None)
        if not execl_kw:
            logger.warning(f"item Missing key __csv__, not stored")
            return item

        execl_kw.setdefault('filename', spider.name)
        async with self.lock:
            await self.save_item(item, **execl_kw)

    async def close_spider(self, spider):
        self.close()


if __name__ == '__main__':
    class TestSpider:
        name = 'TestSpider'


    async def test():
        p = CsvPipeline({})
        await p.process_item({
            'title': '测试',
            'img': '//www.baidu.com/img/flexible/logo/pc/result.png',
            '__csv__': {
                'filename': 'test',
            }
        }, TestSpider())

        await p.close_spider(None)


    asyncio.run(test())



from aioscrapy.exceptions import _InvalidOutput
from aioscrapy.http import Request, Response
from aioscrapy.middleware.absmanager import AbsMiddlewareManager
from aioscrapy.utils.conf import build_component_list
from aioscrapy.utils.tools import call_helper


class DownloaderMiddlewareManager(AbsMiddlewareManager):
    component_name = 'downloader middleware'

    @classmethod
    def _get_mwlist_from_settings(cls, settings):
        return build_component_list(
            settings.getwithbase('DOWNLOADER_MIDDLEWARES'))

    def _add_middleware(self, mw):
        if hasattr(mw, 'process_request'):
            self.methods['process_request'].append(mw.process_request)
        if hasattr(mw, 'process_response'):
            self.methods['process_response'].appendleft(mw.process_response)
        if hasattr(mw, 'process_exception'):
            self.methods['process_exception'].appendleft(mw.process_exception)

    def iter_mw_method(self, spider, process_type: str):
        spider_method = getattr(spider, process_type, None)
        for method in self.methods[process_type]:
            yield method
        if spider_method:
            yield spider_method

    async def process_request(self, spider, request):
        for method in self.iter_mw_method(spider, 'process_request'):
            response = await call_helper(method, request=request, spider=spider)
            if response is not None and not isinstance(response, (Response, Request)):
                raise _InvalidOutput(
                    "Middleware %s.process_request must return None, Response or Request, got %s"
                    % (method.__self__.__class__.__name__, response.__class__.__name__)
                )
            if response:
                return response

    async def process_response(self, spider, request, response):
        if response is None:
            raise TypeError("Received None in process_response")
        elif isinstance(response, Request):
            return response

        for method in self.iter_mw_method(spider, 'process_response'):
            response = await call_helper(method, request=request, response=response, spider=spider)
            if not isinstance(response, (Response, Request)):
                raise _InvalidOutput(
                    "Middleware %s.process_response must return Response or Request, got %s"
                    % (method.__self__.__class__.__name__, type(response))
                )
            if isinstance(response, Request):
                return response
        return response

    async def process_exception(self, spider, request, exception):
        for method in self.iter_mw_method(spider, 'process_exception'):
            response = await call_helper(method, request=request, exception=exception, spider=spider)
            if response is not None and not isinstance(response, (Response, Request)):
                raise _InvalidOutput(
                    "Middleware %s.process_exception must return None, Response or Request, got %s"
                    % (method.__self__.__class__.__name__, type(response))
                )
            if response:
                return response
        return exception



from typing import Literal
from abc import ABCMeta, abstractmethod

from aioscrapy import Request, Spider
from aioscrapy.utils.log import logger


class DupeFilterBase(metaclass=ABCMeta):
    """Request Fingerprint duplicates filter"""

    @classmethod
    @abstractmethod
    def from_crawler(cls, crawler: "aioscrapy.crawler.Crawler"):
        """ Get Instance of RFPDupeFilter from crawler """

    @abstractmethod
    async def request_seen(self, request: Request) -> bool:
        """ Check whether fingerprint of request exists """

    @abstractmethod
    async def close(self, reason: str = '') -> None:
        """ Delete data on close """

    def log(self, request: Request, spider: Spider):
        if self.info:
            logger.info("Filtered duplicate request: %(request)s" % {
                'request': request.meta.get('dupefilter_msg') or request
            })
        elif self.debug:
            logger.debug("Filtered duplicate request: %(request)s" % {
                'request': request.meta.get('dupefilter_msg') or request
            })
        elif self.logdupes:
            msg = ("Filtered duplicate request: %(request)s"
                   " - no more duplicates will be shown"
                   " (see DUPEFILTER_DEBUG to show all duplicates)")
            logger.debug(msg % {'request': request.meta.get('dupefilter_msg') or request})
            self.logdupes = False

        spider.crawler.stats.inc_value('dupefilter/filtered', spider=spider)

    async def done(
            self,
            request: Request,
            done_type: Literal["request_ok", "request_err", "parse_ok", "parse_err"]
    ) -> None:
        """ 根据done_type的状态 控制指纹的移除 """



"""
Extension for collecting core stats like items scraped and start/finish times
"""
from datetime import datetime

from aioscrapy import signals


class CoreStats:

    def __init__(self, stats):
        self.stats = stats
        self.start_time = None

    @classmethod
    def from_crawler(cls, crawler):
        o = cls(crawler.stats)
        crawler.signals.connect(o.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(o.spider_closed, signal=signals.spider_closed)
        crawler.signals.connect(o.item_scraped, signal=signals.item_scraped)
        crawler.signals.connect(o.item_dropped, signal=signals.item_dropped)
        crawler.signals.connect(o.response_received, signal=signals.response_received)
        return o

    def spider_opened(self, spider):
        self.start_time = datetime.now()
        self.stats.set_value('start_time', str(self.start_time), spider=spider)

    def spider_closed(self, spider, reason):
        finish_time = datetime.now()
        elapsed_time = finish_time - self.start_time
        elapsed_time_seconds = elapsed_time.total_seconds()
        self.stats.set_value('elapsed_time_seconds', elapsed_time_seconds, spider=spider)
        self.stats.set_value('finish_time', str(finish_time), spider=spider)
        self.stats.set_value('finish_reason', reason, spider=spider)

    def item_scraped(self, item, spider):
        self.stats.inc_value('item_scraped_count', spider=spider)

    def response_received(self, spider):
        self.stats.inc_value('response_received_count', spider=spider)

    def item_dropped(self, item, spider, exception):
        reason = exception.__class__.__name__
        self.stats.inc_value('item_dropped_count', spider=spider)
        self.stats.inc_value(f'item_dropped_reasons_count/{reason}', spider=spider)



"""
DefaultHeaders downloader middleware

See documentation in docs/topics/downloader-middleware.rst
"""

from aioscrapy.utils.python import without_none_values


class DefaultHeadersMiddleware:

    def __init__(self, headers):
        self._headers = headers

    @classmethod
    def from_crawler(cls, crawler):
        headers = without_none_values(crawler.settings['DEFAULT_REQUEST_HEADERS'])
        return cls(headers.items())

    def process_request(self, request, spider):
        for k, v in self._headers:
            request.headers.setdefault(k, v)



import ssl

import httpx
from httpx import HTTPError as HttpxError

from aioscrapy import Request
from aioscrapy.core.downloader.handlers import BaseDownloadHandler
from aioscrapy.exceptions import DownloadError
from aioscrapy.http import HtmlResponse
from aioscrapy.settings import Settings
from aioscrapy.utils.log import logger


class HttpxDownloadHandler(BaseDownloadHandler):

    def __init__(self, settings):
        self.settings: Settings = settings
        self.httpx_client_session_args: dict = self.settings.get('HTTPX_CLIENT_SESSION_ARGS', {})
        self.verify_ssl: bool = self.settings.get("VERIFY_SSL", True)
        self.ssl_protocol = self.settings.get("SSL_PROTOCOL")  # ssl.PROTOCOL_TLSv1_2
        if self.settings.getbool("FIX_HTTPX_HEADER", True):
            # Fixed non-standard response's header 修复不标准的响应头
            import h11
            import re
            h11._readers.header_field_re = re.compile(b"(?P<field_name>.*?):[ \t](?P<field_value>.*?)")

    @classmethod
    def from_settings(cls, settings: Settings):
        return cls(settings)

    async def download_request(self, request: Request, _) -> HtmlResponse:
        try:
            return await self._download_request(request)
        except HttpxError as e:
            raise DownloadError(real_error=e) from e

    async def _download_request(self, request: Request) -> HtmlResponse:
        kwargs = {
            'timeout': self.settings.get('DOWNLOAD_TIMEOUT'),
            'cookies': dict(request.cookies),
            'data': request.body or None
        }
        headers = request.headers or self.settings.get('DEFAULT_REQUEST_HEADERS')
        kwargs['headers'] = headers

        session_args = self.httpx_client_session_args.copy()
        session_args.setdefault('http2', True)
        session_args.update({
            'verify': request.meta.get('verify_ssl', self.verify_ssl),
            'follow_redirects': self.settings.getbool('REDIRECT_ENABLED', True) if request.meta.get(
                'dont_redirect') is None else request.meta.get('dont_redirect'),
            'max_redirects': self.settings.getint('REDIRECT_MAX_TIMES', 20),
        })
        ssl_ciphers = request.meta.get('TLS_CIPHERS')
        ssl_protocol = request.meta.get('ssl_protocol', self.ssl_protocol)
        if ssl_ciphers or ssl_protocol:
            if ssl_protocol:
                context = ssl.SSLContext(protocol=ssl_protocol)
            else:
                context = ssl.create_default_context()

            ssl_ciphers and context.set_ciphers(ssl_ciphers)
            session_args['verify'] = context

        proxy = request.meta.get("proxy")
        if proxy:
            session_args["proxies"] = proxy
            logger.debug(f"使用代理{proxy}抓取: {request.url}")

        async with httpx.AsyncClient(**session_args) as session:
            response = await session.request(request.method, request.url, **kwargs)
            content = response.read()

        return HtmlResponse(
            str(response.url),
            status=response.status_code,
            headers=response.headers,
            body=content,
            cookies={j.name: j.value or '' for j in response.cookies.jar},
            encoding=response.encoding
        )

    async def close(self):
        pass



import warnings
from functools import wraps

from aioscrapy.exceptions import AioScrapyDeprecationWarning


def deprecated(use_instead=None):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    def deco(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            message = f"Call to deprecated function {func.__name__}."
            if use_instead:
                message += f" Use {use_instead} instead."
            warnings.warn(message, category=AioScrapyDeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)
        return wrapped

    if callable(use_instead):
        deco = deco(use_instead)
        use_instead = None
    return deco



"""
aioscrapy core exceptions

These exceptions are documented in docs/topics/exceptions.rst. Please don't add
new exceptions here without documenting them there.
"""


# Internal


class NotConfigured(Exception):
    """Indicates a missing configuration situation"""
    pass


class _InvalidOutput(TypeError):
    """
    Indicates an invalid value has been returned by a middleware's processing method.
    Internal and undocumented, it should not be raised or caught by user code.
    """
    pass


# HTTP and crawling


class IgnoreRequest(Exception):
    """Indicates a decision was made not to process a request"""


class DontCloseSpider(Exception):
    """Request the spider not to be closed yet"""
    pass


class CloseSpider(Exception):
    """Raise this from callbacks to request the spider to be closed"""

    def __init__(self, reason='cancelled'):
        super().__init__()
        self.reason = reason


class StopDownload(Exception):
    """
    Stop the download of the body for a given response.
    The 'fail' boolean parameter indicates whether or not the resulting partial response
    should be handled by the request errback. Note that 'fail' is a keyword-only argument.
    """

    def __init__(self, *, fail=True):
        super().__init__()
        self.fail = fail


# Items


class DropItem(Exception):
    """Drop item from the item pipeline"""
    pass


class NotSupported(Exception):
    """Indicates a feature or method is not supported"""
    pass


# Commands


class UsageError(Exception):
    """To indicate a command-line usage error"""

    def __init__(self, *a, **kw):
        self.print_help = kw.pop('print_help', True)
        super().__init__(*a, **kw)


class AioScrapyDeprecationWarning(Warning):
    """Warning category for deprecated features, since the default
    DeprecationWarning is silenced on Python 2.7+
    """
    pass


class ContractFail(AssertionError):
    """Error raised in case of a failing contract"""
    pass


class ProxyException(Exception):
    pass


class DownloadError(Exception):
    """下载页面时发生的错误"""

    def __init__(self, *args, real_error=None):
        self.real_error = real_error
        super().__init__(*args)

    def __str__(self):
        if not self.real_error:
            return "DownloadError"

        return f"{self.real_error.__class__.__module__}.{self.real_error.__class__.__name__}: {str(self.real_error)}"


if __name__ == '__main__':
    e = Exception("xxx")
    reason = DownloadError(real_error=e)
    print(reason)
    obj = reason.real_error.__class__
    print(f"{obj.__module__}.{obj.__name__}")



"""Helper functions for working with signals"""
import asyncio

from pydispatch.dispatcher import Anonymous, Any, disconnect, getAllReceivers, liveReceivers
from pydispatch.robustapply import robustApply

from aioscrapy.exceptions import StopDownload
from aioscrapy.utils.log import logger
from aioscrapy.utils.tools import create_task


class _IgnoredException(Exception):
    pass


async def robustApplyWrap(f, recv, *args, **kw):
    dont_log = kw.pop('dont_log', None)
    spider = kw.get('spider', None)
    try:
        result = f(recv, *args, **kw)
        if asyncio.iscoroutine(result):
            return await result
    except (Exception, BaseException) as exc:  # noqa: E722
        if dont_log is None or not isinstance(exc, dont_log):
            logger.exception(f"Error caught on signal handler: {recv}")
        return exc


async def send_catch_log(signal=Any, sender=Anonymous, *arguments, **named):
    """Like pydispatcher.robust.sendRobust but it also logs errors and returns
    Failures instead of exceptions.
    """
    named['dont_log'] = (named.pop('dont_log', _IgnoredException), StopDownload)
    responses = []
    for receiver in liveReceivers(getAllReceivers(sender, signal)):
        result = await robustApplyWrap(robustApply, receiver, signal=signal, sender=sender, *arguments, **named)
        responses.append((receiver, result))
    return responses


async def send_catch_log_deferred(signal=Any, sender=Anonymous, *arguments, **named):
    """Like send_catch_log but supports returning deferreds on signal handlers.
    Returns a deferred that gets fired once all signal handlers deferreds were
    fired.
    """
    dfds = []
    for receiver in liveReceivers(getAllReceivers(sender, signal)):
        dfds.append(
            create_task(
                robustApplyWrap(robustApply, receiver, signal=signal, sender=sender, *arguments, **named)
            )
        )
    return await asyncio.gather(*dfds)


def disconnect_all(signal=Any, sender=Any):
    """Disconnect all signal handlers. Useful for cleaning up after running
    tests
    """
    for receiver in liveReceivers(getAllReceivers(sender, signal)):
        disconnect(receiver, signal=signal, sender=sender)



from aioscrapy.db import db_manager
from aioscrapy.libs.pipelines import DBPipelineBase

from aioscrapy.utils.log import logger


class MongoPipeline(DBPipelineBase):

    def __init__(self, settings, db_type: str):
        super().__init__(settings, db_type)
        self.db_cache = {}
        self.ordered_cache = {}
        self.retry_times = settings.getint("MONGO_TIMEOUT_RETRY_TIMES", 3)

    @classmethod
    def from_settings(cls, settings):
        return cls(settings, 'mongo')

    def parse_item_to_cache(self, item: dict, save_info: dict):
        db_name = save_info.get('db_name')
        table_name = save_info.get('table_name')
        ordered = save_info.get('ordered', False)
        assert table_name is not None, 'please set table_name'
        db_alias = save_info.get('db_alias', ['default'])
        if isinstance(db_alias, str):
            db_alias = [db_alias]

        cache_key = ''.join(db_alias) + (db_name or '') + table_name + str(ordered)

        if self.table_cache.get(cache_key) is None:
            self.db_alias_cache[cache_key] = db_alias
            self.table_cache[cache_key] = table_name
            self.db_cache[cache_key] = db_name
            self.ordered_cache[cache_key] = ordered
            self.item_cache[cache_key] = []

        self.item_cache[cache_key].append(item)
        return cache_key, len(self.item_cache[cache_key])

    async def _save(self, cache_key):
        table_name = self.table_cache[cache_key]
        try:
            for alias in self.db_alias_cache[cache_key]:
                try:
                    executor = db_manager.mongo.executor(alias)
                    result = await executor.insert(
                        table_name, self.item_cache[cache_key], db_name=self.db_cache[cache_key],
                        ordered=self.ordered_cache[cache_key], retry_times=self.retry_times
                    )
                    logger.info(
                        f'table:{alias}->{table_name} sum:{len(self.item_cache[cache_key])} ok:{len(result.inserted_ids)}'
                    )
                except Exception as e:
                    logger.exception(f'save data error, table:{alias}->{table_name}, err_msg:{e}')
        finally:
            self.item_cache[cache_key] = []



import aioscrapy
from aioscrapy.commands import AioScrapyCommand


class Command(AioScrapyCommand):
    default_settings = {
        'LOG_ENABLED': False,
        'SPIDER_LOADER_WARN_ONLY': True
    }

    def syntax(self):
        return "[-v]"

    def short_desc(self):
        return "Print aioscrapy version"

    def add_options(self, parser):
        AioScrapyCommand.add_options(self, parser)

    def run(self, args, opts):
        print(f"aioscrapy {aioscrapy.__version__}")



from importlib import import_module
from typing import Any

import aioscrapy
from aioscrapy.db.absmanager import AbsDBPoolManager
from aioscrapy.db.aioredis import redis_manager
from aioscrapy.utils.log import logger
from aioscrapy.utils.misc import load_object

__all__ = ['db_manager', 'get_pool', 'get_manager']

DB_MODULE_MAP = {
    'redis': ('redis', 'aioscrapy.db.aioredis.redis_manager'),
    'aiomysql': ('mysql', 'aioscrapy.db.aiomysql.mysql_manager'),
    'aio_pika': ('rabbitmq', 'aioscrapy.db.aiorabbitmq.rabbitmq_manager'),
    'motor': ('mongo', 'aioscrapy.db.aiomongo.mongo_manager'),
    'asyncpg': ('pg', 'aioscrapy.db.aiopg.pg_manager'),
}

db_manager_map = {}

for module_name, (manager_key, class_path) in DB_MODULE_MAP.items():
    try:
        import_module(module_name)
    except ImportError:
        pass
    else:
        db_manager_map[manager_key] = load_object(class_path)


class DBManager:

    @staticmethod
    def get_manager(db_type: str) -> AbsDBPoolManager:
        manager = db_manager_map.get(db_type)
        assert manager is not None, f"Not support db type：{db_type}"
        return manager

    def get_pool(self, db_type: str, alias='default') -> Any:
        manager = self.get_manager(db_type)
        return manager.get_pool(alias)

    @staticmethod
    async def close_all() -> None:
        for manager in db_manager_map.values():
            await manager.close_all()

    @staticmethod
    async def from_dict(db_args: dict) -> None:
        for db_type, args in db_args.items():
            manager = db_manager_map.get(db_type)
            if manager is None:
                logger.warning(f'Not support db type: {db_type}; Only {", ".join(db_manager_map.keys())} supported')
            await manager.from_dict(args)

    @staticmethod
    async def from_settings(settings: aioscrapy.Settings) -> None:
        for manager in db_manager_map.values():
            await manager.from_settings(settings)

    async def from_crawler(self, crawler: "aioscrapy.Crawler") -> None:
        return await self.from_settings(crawler.settings)

    def __getattr__(self, db_type: str) -> Any:
        if db_type not in db_manager_map:
            raise AttributeError(f'Not support db type: {db_type}')
        return db_manager_map[db_type]


db_manager = DBManager()
get_manager = db_manager.get_manager
get_pool = db_manager.get_pool



import os
from typing import Optional, Set

from aioscrapy import Request
from aioscrapy.dupefilters import DupeFilterBase


class DiskRFPDupeFilter(DupeFilterBase):
    """Request Fingerprint duplicates filter built with Disk storage"""

    def __init__(self, path: Optional[str] = None, debug: bool = False, info: bool = False):
        self.file: Optional["File object"] = None
        self.debug = debug
        self.fingerprints: Set = set()
        self.logdupes: bool = True
        self.info: bool = info
        if path:
            self.file = open(os.path.join(path, 'requests.seen'), 'a+')
            self.file.seek(0)
            self.fingerprints.update(x.rstrip() for x in self.file)

    @classmethod
    def from_crawler(cls, crawler: "aioscrapy.crawler.Crawler"):
        debug = crawler.settings.getbool('DUPEFILTER_DEBUG')
        info = crawler.settings.getbool('DUPEFILTER_INFO')
        path = crawler.settings.get('JOBDIR', './job_dir')
        if path and not os.path.exists(path):
            os.makedirs(path)
        return cls(path, debug, info)

    async def request_seen(self, request: Request) -> bool:
        if request.fingerprint in self.fingerprints:
            return True
        self.fingerprints.add(request.fingerprint)
        if self.file:
            self.file.write(request.fingerprint + '\n')
        return False

    async def close(self, reason: str = '') -> None:
        if self.file:
            self.file.close()


RFPDupeFilter = DiskRFPDupeFilter



from aioscrapy.middleware.absmanager import AbsMiddlewareManager
from aioscrapy.utils.conf import build_component_list


class ItemPipelineManager(AbsMiddlewareManager):
    component_name = 'item pipeline'

    @classmethod
    def _get_mwlist_from_settings(cls, settings):
        return build_component_list(settings.getwithbase('ITEM_PIPELINES'))

    def _add_middleware(self, pipe):
        super()._add_middleware(pipe)
        if hasattr(pipe, 'process_item'):
            self.methods['process_item'].append(pipe.process_item)

    async def process_item(self, item, spider):
        return await self._process_chain('process_item', item, spider)



"""
An extension to retry failed requests that are potentially caused by temporary
problems such as a connection timeout or HTTP 500 error.

You can change the behaviour of this middleware by modifing the scraping settings:
RETRY_TIMES - how many times to retry a failed page
RETRY_HTTP_CODES - which HTTP response codes to retry

Failed pages are collected on the scraping process and rescheduled at the end,
once the spider has finished crawling all regular (non failed) pages.
"""
from typing import Optional, Union

from anyio import EndOfStream

try:
    from asyncio.exceptions import TimeoutError
except:
    from concurrent.futures._base import TimeoutError

from aioscrapy.exceptions import ProxyException, DownloadError, NotConfigured
from aioscrapy.http.request import Request
from aioscrapy.spiders import Spider
from aioscrapy.utils.log import logger as retry_logger
from aioscrapy.utils.python import global_object_name

NEED_RETRY_ERROR = (TimeoutError, ConnectionRefusedError, IOError, ProxyException, DownloadError, EndOfStream)


def get_retry_request(
        request: Request,
        *,
        spider: Spider,
        reason: Union[str, Exception] = 'unspecified',
        max_retry_times: Optional[int] = None,
        priority_adjust: Optional[int] = None,
        logger=retry_logger,
        stats_base_key: str = 'retry',
):
    """
    使用了scrapy的retry，将日志等级改为info
    """
    settings = spider.crawler.settings
    stats = spider.crawler.stats
    retry_times = request.meta.get('retry_times', 0) + 1
    if max_retry_times is None:
        max_retry_times = request.meta.get('max_retry_times')
        if max_retry_times is None:
            max_retry_times = settings.getint('RETRY_TIMES')
    if retry_times <= max_retry_times:
        if callable(reason):
            reason = reason()
        if isinstance(reason, Exception):
            reason = global_object_name((getattr(reason, "real_error", None) or reason).__class__)

        logger.info(
            "Retrying %(request)s (failed %(retry_times)d times): %(reason)s" % {
                'request': request, 'retry_times': retry_times, 'reason': reason
            },
        )
        new_request = request.copy()
        new_request.meta['retry_times'] = retry_times
        new_request.dont_filter = True
        if priority_adjust is None:
            priority_adjust = settings.getint('RETRY_PRIORITY_ADJUST')
        new_request.priority = request.priority + priority_adjust

        stats.inc_value(f'{stats_base_key}/count')
        stats.inc_value(f'{stats_base_key}/reason_count/{reason}')
        return new_request
    else:
        stats.inc_value(f'{stats_base_key}/max_reached')
        logger.error(
            "Gave up retrying %(request)s (failed %(retry_times)d times): "
            "%(reason)s" % {'request': request, 'retry_times': retry_times, 'reason': reason}
        )
        return None


class RetryMiddleware:
    EXCEPTIONS_TO_RETRY = NEED_RETRY_ERROR

    def __init__(self, settings):
        if not settings.getbool('RETRY_ENABLED'):
            raise NotConfigured
        self.max_retry_times = settings.getint('RETRY_TIMES')
        self.retry_http_codes = set(int(x) for x in settings.getlist('RETRY_HTTP_CODES'))
        self.priority_adjust = settings.getint('RETRY_PRIORITY_ADJUST')

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings)

    def process_response(self, request, response, spider):
        if request.meta.get('dont_retry', False):
            return response
        if response.status in self.retry_http_codes:
            reason = f"Retry response status code： {response.status}"
            return self._retry(request, reason, spider) or response
        return response

    def process_exception(self, request, exception, spider):
        if (
                isinstance(exception, self.EXCEPTIONS_TO_RETRY)
                and not request.meta.get('dont_retry', False)
        ):
            return self._retry(request, exception, spider)

    def _retry(self, request, reason, spider):
        max_retry_times = request.meta.get('max_retry_times', self.max_retry_times)
        priority_adjust = request.meta.get('priority_adjust', self.priority_adjust)
        return get_retry_request(
            request,
            reason=reason,
            spider=spider,
            max_retry_times=max_retry_times,
            priority_adjust=priority_adjust,
        )



from contextlib import asynccontextmanager

from asyncpg.pool import create_pool

import aioscrapy
from aioscrapy.db.absmanager import AbsDBPoolManager


class PGExecutor:
    def __init__(self, alias: str, pool_manager: "AioPGPoolManager"):
        self.alias = alias
        self.pool_manager = pool_manager

    async def insert(self, sql, value):
        async with self.pool_manager.get(self.alias) as connect:
            try:
                result = await connect.executemany(sql, value)
                return result
            except Exception as e:
                await connect.rollback()
                raise Exception from e

    async def fetch(self, sql: str):
        async with self.pool_manager.get(self.alias) as connect:
            return await connect.fetch(sql)

    async def query(self, sql: str):
        return await self.fetch(sql)


class AioPGPoolManager(AbsDBPoolManager):
    _clients = {}

    async def create(self, alias: str, params: dict):
        if alias in self._clients:
            return self._clients[alias]

        params = params.copy()
        params.setdefault('timeout', 30)
        pg_pool = await create_pool(**params)
        return self._clients.setdefault(alias, pg_pool)

    def get_pool(self, alias: str):
        pg_pool = self._clients.get(alias)
        assert pg_pool is not None, f"Dont create the PG pool named {alias}"
        return pg_pool

    @asynccontextmanager
    async def get(self, alias: str):
        """ Get connection of pg """
        pg_pool = self.get_pool(alias)
        conn = await pg_pool.acquire()
        try:
            yield conn
        finally:
            await pg_pool.release(conn)

    def executor(self, alias: str) -> PGExecutor:
        return PGExecutor(alias, self)

    async def close(self, alias: str):
        pg_pool = self._clients.pop(alias, None)
        if pg_pool:
            await pg_pool.close()

    async def close_all(self):
        for alias in list(self._clients.keys()):
            await self.close(alias)

    async def from_dict(self, db_args: dict):
        for alias, pg_args in db_args.items():
            await self.create(alias, pg_args)

    async def from_settings(self, settings: aioscrapy.Settings):
        for alias, pg_args in settings.getdict('PG_ARGS').items():
            await self.create(alias, pg_args)


pg_manager = AioPGPoolManager()

if __name__ == '__main__':
    import asyncio


    async def test():
        pg_pool = await pg_manager.create(
            'default',
            dict(
                user='username',
                password='pwd',
                database='dbname',
                host='127.0.0.1'
            )
        )

        # 方式一:
        conn = await pg_pool.acquire()
        try:
            result = await conn.fetch('select 1 ')
            print(tuple(result[0]))
        finally:
            await pg_pool.release(conn)

        # 方式二:
        async with pg_manager.get('default') as conn:
            result = await conn.fetch('select 1 ')
            print(tuple(result[0]))


    asyncio.run(test())



import asyncio
import pprint
import signal
import sys
import warnings
from typing import Optional, Type, Union, Any

from zope.interface.exceptions import DoesNotImplement

try:
    # zope >= 5.0 only supports MultipleInvalid
    from zope.interface.exceptions import MultipleInvalid
except ImportError:
    MultipleInvalid = None

from zope.interface.verify import verifyClass
from aioscrapy.logformatter import LogFormatter
from aioscrapy import Spider
from aioscrapy.settings import overridden_settings, Settings
from aioscrapy.utils.log import configure_logging, logger
from aioscrapy.utils.misc import load_object, load_instance
from aioscrapy.spiderloader import ISpiderLoader
from aioscrapy.exceptions import AioScrapyDeprecationWarning
from aioscrapy.db import db_manager

from aioscrapy.utils.tools import async_generator_wrapper
from aioscrapy.middleware import ExtensionManager
from aioscrapy.core.engine import ExecutionEngine
from aioscrapy.signalmanager import SignalManager
from aioscrapy.utils.ossignal import install_shutdown_handlers, signal_names
from aioscrapy.statscollectors import StatsCollector


class Crawler:

    def __init__(self, spidercls: Type[Spider], settings: Union[Settings, dict, None] = None) -> None:

        if isinstance(spidercls, Spider):
            raise ValueError('The spidercls argument must be a class, not an object')

        if isinstance(settings, dict) or settings is None:
            settings = Settings(settings)

        self.spidercls = spidercls
        self.settings = settings.copy()
        self.spidercls.update_settings(self.settings)

        self.settings.freeze()
        self.signals: Optional[SignalManager] = None
        self.stats: Optional[StatsCollector] = None
        self.crawling = False
        self.spider: Optional[Spider] = None
        self.engine: Optional[ExecutionEngine] = None
        self.extensions: Optional[ExtensionManager] = None
        self.logformatter: Optional[LogFormatter] = None

    async def crawl(self, *args, **kwargs) -> None:
        try:
            configure_logging(self.spidercls, self.settings)

            if self.crawling:
                raise RuntimeError("Crawling already taking place")

            self.crawling = True
            self.signals = SignalManager(self)
            self.stats = load_object(self.settings['STATS_CLASS'])(self)

            logger.info(f"Overridden settings:\n{pprint.pformat(dict(overridden_settings(self.settings)))}")

            self.spider = await self.spidercls.from_crawler(self, *args, **kwargs)
            self.spider.stats = self.stats
            self.logformatter = await load_instance(self.settings['LOG_FORMATTER'], crawler=self)
            self.extensions = await ExtensionManager.from_crawler(self)
            self.engine = ExecutionEngine(self)
            # 创建所有数据库链接
            await db_manager.from_crawler(self)
            start_requests = await async_generator_wrapper(self.spider.start_requests())
            await self.engine.start(self.spider, start_requests)
        except Exception as e:
            logger.exception(e)
            self.crawling = False
            if self.engine is not None:
                await self.engine.close()
            raise e

    async def stop(self, signum=None) -> None:
        """Starts a graceful stop of the crawler and returns a deferred that is
        fired when the crawler is stopped."""
        if signum is not None:
            asyncio.current_task().set_name(self.spidercls.name)
            logger.info(
                "Received  %(signame)s, shutting down gracefully. Send again to force" % {
                    'signame': signal_names[signum]
                }
            )
        if self.crawling:
            self.crawling = False
            await self.engine.stop()

    def _signal_shutdown(self, signum: Any, _) -> None:
        asyncio.create_task(self.stop(signum))


class CrawlerRunner:
    crawlers = property(
        lambda self: self._crawlers,
        doc="Set of :class:`crawlers <scrapy.crawler.Crawler>` started by "
            ":meth:`crawl` and managed by this class."
    )

    @staticmethod
    def _get_spider_loader(settings: Settings) -> ISpiderLoader:
        """ Get SpiderLoader instance from settings """
        cls_path = settings.get('SPIDER_LOADER_CLASS')
        loader_cls = load_object(cls_path)
        excs = (DoesNotImplement, MultipleInvalid) if MultipleInvalid else DoesNotImplement
        try:
            verifyClass(ISpiderLoader, loader_cls)
        except excs as e:
            warnings.warn(
                'SPIDER_LOADER_CLASS (previously named SPIDER_MANAGER_CLASS) does '
                'not fully implement scrapy.interfaces.ISpiderLoader interface. '
                'Please add all missing methods to avoid unexpected runtime errors.',
                category=AioScrapyDeprecationWarning, stacklevel=2
            )
        return loader_cls.from_settings(settings.frozencopy())

    def __init__(self, settings: Union[Settings, dict, None] = None) -> None:
        if isinstance(settings, dict) or settings is None:
            settings = Settings(settings)
        self.settings = settings
        self.spider_loader = self._get_spider_loader(settings)
        self._crawlers = {}
        self._active = set()
        self.bootstrap_failed = False

    @property
    def spiders(self):
        warnings.warn("CrawlerRunner.spiders attribute is renamed to "
                      "CrawlerRunner.spider_loader.",
                      category=AioScrapyDeprecationWarning, stacklevel=2)
        return self.spider_loader

    def crawl_soon(
            self,
            crawler_or_spidercls: Union[Type[Spider], Crawler],
            *args,
            settings: Union[Settings, dict, None] = None,
            **kwargs
    ) -> None:
        crawler = self.crawl(crawler_or_spidercls, settings=settings)
        self.crawlers.setdefault(crawler, (args, kwargs))
        self.active_crawler(crawler, *args, **kwargs)

    def active_crawler(self, crawler: Crawler, *args, **kwargs) -> None:
        task = asyncio.create_task(crawler.crawl(*args, **kwargs), name=crawler.spidercls.name)
        self._active.add(task)

        def _done(result):
            self.crawlers.pop(crawler, None)
            self._active.discard(task)
            self.bootstrap_failed |= not getattr(crawler, 'spider', None)
            return result

        task.add_done_callback(_done)

    def crawl(
            self,
            crawler_or_spidercls: Union[Type[Spider], Crawler],
            *args,
            settings: Union[Settings, dict, None] = None,
            **kwargs
    ) -> Crawler:
        if isinstance(crawler_or_spidercls, Spider):
            raise ValueError(
                'The crawler_or_spidercls argument cannot be a spider object, '
                'it must be a spider class (or a Crawler object)')
        crawler = self.create_crawler(crawler_or_spidercls, settings or self.settings)
        self.crawlers.setdefault(crawler, (args, kwargs))
        return crawler

    def create_crawler(
            self,
            crawler_or_spidercls: Union[Type[Spider], Crawler, str],
            settings: Union[Settings, dict, None]
    ) -> Crawler:
        if isinstance(crawler_or_spidercls, Spider):
            raise ValueError(
                'The crawler_or_spidercls argument cannot be a spider object, '
                'it must be a spider class (or a Crawler object)')
        if isinstance(crawler_or_spidercls, Crawler):
            return crawler_or_spidercls
        return self._create_crawler(crawler_or_spidercls, settings)

    def _create_crawler(
            self,
            spidercls: Union[Type[Spider], str],
            settings: Union[Settings, dict, None]
    ) -> Crawler:
        if isinstance(spidercls, str):
            spidercls = self.spider_loader.load(spidercls)
        return Crawler(spidercls, settings=settings)

    async def stop(self, signum=None) -> None:
        await asyncio.gather(*[c.stop(signum) for c in self.crawlers])


class CrawlerProcess(CrawlerRunner):

    def __init__(
            self,
            settings: Union[Settings, dict, None] = None,
            install_root_handler: bool = True
    ) -> None:
        super().__init__(settings)
        install_shutdown_handlers(self._signal_shutdown)

    def _signal_shutdown(self, signum: Any, _) -> None:
        install_shutdown_handlers(self._signal_kill)
        asyncio.create_task(self.stop(signum))

    def _signal_kill(self, signum: Any, _) -> None:
        install_shutdown_handlers(signal.SIG_IGN)
        signame = signal_names[signum]
        logger.info('Received %(signame)s twice, forcing unclean shutdown' % {'signame': signame})
        asyncio.create_task(self._stop_reactor())

    async def run(self) -> None:
        try:
            for crawler, (args, kwargs) in self.crawlers.items():
                self.active_crawler(crawler, *args, **kwargs)
            while self._active:
                await asyncio.gather(*self._active)
        finally:
            await self.recycle_db_connect()

    def start(self, use_windows_selector_eventLoop: bool = False) -> None:
        if sys.platform.startswith('win'):
            if use_windows_selector_eventLoop:
                asyncio.set_event_loop_policy(asyncio.windows_events.WindowsSelectorEventLoopPolicy())
            else:
                asyncio.set_event_loop(asyncio.windows_events.ProactorEventLoop())
        else:
            try:
                import uvloop
                asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
            except ImportError:
                pass
        asyncio.run(self.run())

    async def _stop_reactor(self) -> None:
        try:
            await self.recycle_db_connect()
        finally:
            asyncio.get_event_loop().stop()

    async def recycle_db_connect(self) -> None:
        # recycle pool of db_manager
        if not len(self._active):
            await db_manager.close_all()



from aioscrapy.db import db_manager
from aioscrapy.libs.pipelines import DBPipelineBase

from aioscrapy.utils.log import logger


class PGPipeline(DBPipelineBase):

    @classmethod
    def from_settings(cls, settings):
        return cls(settings, 'pg')

    async def _save(self, cache_key):
        table_name = self.table_cache[cache_key]
        try:
            for alias in self.db_alias_cache[cache_key]:
                async with db_manager.pg.get(alias) as conn:
                    try:
                        num = await conn.executemany(
                            self.insert_sql_cache[cache_key], self.item_cache[cache_key]
                        )
                        logger.info(f'table:{alias}->{table_name} sum:{len(self.item_cache[cache_key])} ok:{num}')
                    except Exception as e:
                        logger.exception(f'save data error, table:{alias}->{table_name}, err_msg:{e}')
        finally:
            self.item_cache[cache_key] = []



import asyncio
import os
import platform
import random
import time

from aiohttp import ClientSession

from aioscrapy import Settings
from aioscrapy import signals
from aioscrapy.utils.log import _logger, logger
from aioscrapy.utils.tools import create_task


class InfluxBase:
    @staticmethod
    def format_metric(metric_name, value, spider_name, location, measurement=None):
        measurement = measurement or metric_name
        return f"{measurement},spider_name={spider_name},location={location} {metric_name}={value} {time.time_ns() + int(random.random() * 100000)}"

    async def record(self, obj: "Metric"):
        raise NotImplementedError

    async def close(self):
        pass


class InfluxHttp(InfluxBase):
    def __init__(self, spider_name: str, settings: Settings):
        influxdb_url = settings.get('METRIC_INFLUXDB_URL')
        token = settings.get('METRIC_INFLUXDB_TOKEN')
        location = settings.get('METRIC_LOCATION')
        self.retry_times = settings.getint('METRIC_RETRY_TIMES', 5)
        self.location = location or f"{platform.node()}_{os.getpid()}"
        self.spider_name = spider_name
        self.session = ClientSession(headers={
            "Authorization": f"Token {token}",
            "Content-Type": "text/plain; charset=utf-8",
            "Accept": "application/json",
        })
        self.url = influxdb_url
        self.lock = asyncio.Lock()

    async def emit(self, data):
        async with self.session.post(self.url, data=data) as response:
            await response.read()
            logger.debug(f"emit metric success<{response.status}>: \n{data}")

    async def record(self, obj: "Metric"):
        async with self.lock:
            data = ''
            for metric_name in obj.metrics.keys():
                current_cnt = obj.stats.get_value(metric_name, 0)
                if not isinstance(current_cnt, (int, float)):
                    continue
                cnt = current_cnt - obj.prev.get(metric_name, 0)
                if cnt:
                    data += self.format_metric(
                        metric_name.replace('/', '-'), cnt, self.spider_name, self.location
                    ) + '\n'
                obj.prev[metric_name] = current_cnt
            if data:
                for _ in range(self.retry_times):
                    try:
                        await self.emit(data)
                        return
                    except:
                        continue
                logger.warning(f"emit metric failed:\n{data}")

    async def close(self):
        if self.session is not None:
            await self.session.close()
            await asyncio.sleep(0.250)


class InfluxLog(InfluxBase):
    def __init__(self, spider_name: str, settings: Settings):
        location = settings.get('METRIC_LOCATION')
        self.location = location or f"{platform.node()}_{os.getpid()}"
        self.spider_name = spider_name

        log_args = settings.getdict('METRIC_LOG_ARGS')
        if log_args:
            log_args.update(dict(
                filter=lambda record: record["extra"].get("metric") is not None,
                encoding="utf-8"
            ))
            for k, v in dict(
                    sink=f'{spider_name}.metric', level="INFO", rotation='20MB', retention=3,
                    format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> <level>{message}</level>",
            ).items():
                log_args.setdefault(k, v)

            _logger.add(**log_args)
            self.log = _logger.bind(metric="metric")
        else:
            self.log = logger

    async def record(self, obj: "Metric"):
        for metric_name in obj.metrics.keys():
            current_cnt = obj.stats.get_value(metric_name, 0)
            if not isinstance(current_cnt, (int, float)):
                continue
            prev_cnt = obj.prev.get(metric_name, 0)
            cnt = current_cnt - prev_cnt
            if cnt:
                msg = self.format_metric(metric_name.replace('/', '-'), cnt, self.spider_name, self.location)
                self.log.info(f'METRIC: {msg}')
            obj.prev[metric_name] = current_cnt


class Metric:
    """Log Metric scraping stats periodically"""

    def __init__(self, stats, spider_name, settings, interval=10.0):
        if settings.get('METRIC_INFLUXDB_URL'):
            self.influx = InfluxHttp(spider_name, settings)
        else:
            self.influx = InfluxLog(spider_name, settings)
        self.stats = stats
        self.metrics = settings.getdict('METRICS') or self.stats._stats
        self.interval = interval
        self.task = None
        self.prev = {}

    @classmethod
    def from_crawler(cls, crawler):
        interval = crawler.settings.getfloat('METRIC_INTERVAL', 10.0)
        o = cls(crawler.stats, crawler.spider.name, crawler.settings, interval)
        crawler.signals.connect(o.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(o.spider_closed, signal=signals.spider_closed)
        return o

    def spider_opened(self, spider):
        self.task = create_task(self.run(spider))

    async def run(self, spider):
        await asyncio.sleep(self.interval)
        await self.influx.record(self)
        self.task = create_task(self.run(spider))

    async def spider_closed(self, spider, reason):
        if self.task and not self.task.done():
            self.task.cancel()
        await self.influx.record(self)
        await self.influx.close()



import asyncio
import time
from typing import Optional

from aioscrapy.db import db_manager
from aioscrapy.exceptions import ProxyException
from aioscrapy.proxy import AbsProxy
from aioscrapy.utils.log import logger
from aioscrapy.utils.tools import create_task


class RedisProxy(AbsProxy):
    def __init__(
            self,
            settings,
            crawler,
            proxy_queue: Optional["Redis"] = None,
            proxy_key: Optional[str] = None
    ):
        super().__init__(settings)
        self.crawler = crawler
        self.proxy_queue = proxy_queue
        self.proxy_key = proxy_key
        self.lock = asyncio.Lock()

    @classmethod
    async def from_crawler(cls, crawler) -> "RedisProxy":
        settings = crawler.settings
        proxy_key = settings.get('PROXY_KEY')
        assert proxy_key is not None, "Not configured：'PROXY_KEY'"
        alias = settings.get("PROXY_QUEUE_ALIAS", 'proxy')
        proxy_queue = db_manager.redis(alias)
        return cls(
            settings,
            crawler,
            proxy_queue=proxy_queue,
            proxy_key=proxy_key
        )

    async def fill_proxy(self, redis_key: str, count: int) -> None:
        script = f"""
            local redis_key = KEYS[1]
            local min_score = ARGV[1]
            local max_score = ARGV[2]
            local num = redis.call('ZCOUNT', redis_key, min_score, max_score)
            local start = 0
            if num>{count} then
                math.randomseed({time.time() * 1000})
                start = math.random(0, num-{count})
            end
            return redis.call('ZRANGEBYSCORE', redis_key, min_score, max_score, 'LIMIT', start, {count})
        """
        cmd_script = self.proxy_queue.register_script(script)
        result = await cmd_script(keys=[redis_key], args=[100, 100])
        if not result:
            result = await cmd_script(keys=[redis_key], args=[0, 100])
        proxies = [ip.decode() if ip.decode().startswith('http') else f'http://{ip.decode()}' for ip in result]
        self.cache.extend(proxies)
        logger.info(f'Get proxy from redis: {proxies}')

    async def get(self) -> str:
        if len(self.cache) < self.min_count:
            async with self.lock:
                len(self.cache) < self.min_count and await self.fill_proxy(self.proxy_key, self.max_count - len(self.cache))
        try:
            proxy = self.cache.pop(0)
            self.cache.append(proxy)
            return proxy
        except IndexError:
            logger.warning("Not available proxy, Closing spider")
            create_task(self.crawler.engine.stop(reason="Not available proxy"))
            raise ProxyException("Not available proxy")



"""
aioscrapy - a web crawling and web scraping framework written for Python
"""

import pkgutil
import sys

# Declare top-level shortcuts
from aioscrapy.spiders import Spider
from aioscrapy.http import Request, FormRequest
from aioscrapy.settings import Settings
from aioscrapy.crawler import Crawler
from aioscrapy.utils.log import logger


__all__ = [
    '__version__', 'version_info', 'Spider', 'Request', 'FormRequest', 'Crawler', 'Settings', 'logger'
]


# aioscrapy versions
__version__ = (pkgutil.get_data(__package__, "VERSION") or b"").decode("ascii").strip()
version_info = tuple(int(v) if v.isdigit() else v for v in __version__.split('.'))


# Check minimum required Python version
if sys.version_info < (3, 9):
    print("aioscrapy %s requires Python 3.9+" % __version__)
    sys.exit(1)


del pkgutil
del sys



"""
Depth Spider Middleware

See documentation in docs/topics/spider-middleware.rst
"""

from aioscrapy.http import Request

from aioscrapy.utils.log import logger


class DepthMiddleware:

    def __init__(self, maxdepth, stats, verbose_stats=False, prio=1):
        self.maxdepth = maxdepth
        self.stats = stats
        self.verbose_stats = verbose_stats
        self.prio = prio

    @classmethod
    def from_crawler(cls, crawler):
        settings = crawler.settings
        maxdepth = settings.getint('DEPTH_LIMIT')
        verbose = settings.getbool('DEPTH_STATS_VERBOSE')
        prio = settings.getint('DEPTH_PRIORITY')
        return cls(maxdepth, crawler.stats, verbose, prio)

    async def process_spider_output(self, response, result, spider):
        def _filter(request):
            if isinstance(request, Request):
                depth = response.meta['depth'] + 1
                request.meta['depth'] = depth
                if self.prio:
                    request.priority -= depth * self.prio
                if self.maxdepth and depth > self.maxdepth:
                    logger.debug("Ignoring link (depth > %(maxdepth)d): %(requrl)s " % {
                        'maxdepth': self.maxdepth, 'requrl': request.url
                    })
                    return False
                else:
                    if self.verbose_stats:
                        self.stats.inc_value(f'request_depth_count/{depth}',
                                             spider=spider)
                    self.stats.max_value('request_depth_max', depth,
                                         spider=spider)
            return True

        # base case (depth=0)
        if 'depth' not in response.meta:
            response.meta['depth'] = 0
            if self.verbose_stats:
                self.stats.inc_value('request_depth_count/0', spider=spider)

        return (r async for r in result or () if _filter(r))



import asyncio
import re
import ssl
from typing import Optional

import aiohttp
from aiohttp.client_exceptions import ClientError

from aioscrapy import Request
from aioscrapy.core.downloader.handlers import BaseDownloadHandler
from aioscrapy.exceptions import DownloadError
from aioscrapy.http import HtmlResponse
from aioscrapy.settings import Settings
from aioscrapy.utils.log import logger


class AioHttpDownloadHandler(BaseDownloadHandler):
    session: Optional[aiohttp.ClientSession] = None

    def __init__(self, settings: Settings):
        self.settings = settings
        self.aiohttp_client_session_args: dict = settings.getdict('AIOHTTP_CLIENT_SESSION_ARGS')
        self.verify_ssl: Optional[bool] = settings.get("VERIFY_SSL")
        self.ssl_protocol = settings.get("SSL_PROTOCOL")  # ssl.PROTOCOL_TLSv1_2
        self.use_session: bool = settings.getbool("USE_SESSION", False)

    @classmethod
    def from_settings(cls, settings: Settings):
        return cls(settings)

    def get_session(self, *args, **kwargs) -> aiohttp.ClientSession:
        if self.session is None:
            self.session = aiohttp.ClientSession(*args, **kwargs)
        return self.session

    async def download_request(self, request: Request, _) -> HtmlResponse:
        try:
            return await self._download_request(request)
        except ClientError as e:
            raise DownloadError(real_error=e) from e

    async def _download_request(self, request: Request) -> HtmlResponse:
        kwargs = {
            'verify_ssl': request.meta.get('verify_ssl', self.verify_ssl),
            'timeout': request.meta.get('download_timeout', 180),
            'cookies': dict(request.cookies),
            'data': request.body or None,
            'allow_redirects': self.settings.getbool('REDIRECT_ENABLED', True) if request.meta.get(
                'dont_redirect') is None else request.meta.get('dont_redirect'),
            'max_redirects': self.settings.getint('REDIRECT_MAX_TIMES', 20),
        }

        headers = request.headers or self.settings.get('DEFAULT_REQUEST_HEADERS')
        kwargs['headers'] = headers

        ssl_ciphers: str = request.meta.get('TLS_CIPHERS')
        ssl_protocol = request.meta.get('ssl_protocol', self.ssl_protocol)
        if ssl_ciphers or ssl_protocol:
            if ssl_protocol:
                context = ssl.SSLContext(protocol=ssl_protocol)
            else:
                context = ssl.create_default_context()

            ssl_ciphers and context.set_ciphers(ssl_ciphers)
            kwargs['ssl'] = context
            kwargs['verify_ssl'] = True

        proxy: str = request.meta.get("proxy")
        if proxy:
            kwargs["proxy"] = proxy
            logger.debug(f"使用代理{proxy}抓取: {request.url}")

        if self.use_session:
            # Not recommended to use session, The abnormal phenomena will occurs when using tunnel proxy
            session = self.get_session(**self.aiohttp_client_session_args)
            async with session.request(request.method, request.url, **kwargs) as response:
                content: bytes = await response.read()

        else:
            async with aiohttp.ClientSession(**self.aiohttp_client_session_args) as session:
                async with session.request(request.method, request.url, **kwargs) as response:
                    content: bytes = await response.read()

        r_cookies = response.cookies.output() or None
        if r_cookies:
            r_cookies = {
                cookie[0]: cookie[1] for cookie in re.findall(r'Set-Cookie: (.*?)=(.*?); Domain', r_cookies, re.S)
            }

        return HtmlResponse(
            str(response.url),
            status=response.status,
            headers=response.headers,
            body=content,
            cookies=r_cookies,
            encoding=response.charset
        )

    async def close(self):
        if self.session is not None:
            await self.session.close()

            # Wait 250 ms for the underlying SSL connections to close
            # https://docs.aiohttp.org/en/latest/client_advanced.html#graceful-shutdown
            await asyncio.sleep(0.250)



from contextlib import asynccontextmanager

import aio_pika
from aio_pika.exceptions import QueueEmpty
from aio_pika.pool import Pool

import aioscrapy
from aioscrapy.db.absmanager import AbsDBPoolManager


class RabbitmqExecutor:
    def __init__(self, alias: str, pool_manager: "AioRabbitmqManager"):
        self.alias = alias
        self.pool_manager = pool_manager

    async def clean_message_queue(
            self,
            routing_key: str,
            *args,
            if_unused=False,
            if_empty=False,
            timeout=None,
            **kwargs
    ):
        async with self.pool_manager.get(self.alias) as channel:
            queue = await channel.declare_queue(routing_key, *args, **kwargs)
            await queue.delete(if_unused=if_unused, if_empty=if_empty, timeout=timeout)

    async def get_message_count(
            self,
            routing_key: str,
            *args,
            **kwargs
    ):
        async with self.pool_manager.get(self.alias) as channel:
            queue = await channel.declare_queue(routing_key, *args, **kwargs)
            try:
                result = await queue.get(no_ack=False)
                await result.nack()
                return result.message_count
            except QueueEmpty:
                return 0

    async def get_message(
            self,
            routing_key: str,
            *args,
            **kwargs
    ):
        async with self.pool_manager.get(self.alias) as channel:
            queue = await channel.declare_queue(routing_key, *args, **kwargs)
            try:
                result = await queue.get(no_ack=True)
                return result.body
            except QueueEmpty:
                return None

    async def publish(
            self,
            routing_key: str,
            body: bytes,
            *args,
            mandatory=True,
            immediate=False,
            timeout=None,
            **kwargs
    ):
        async with self.pool_manager.get(self.alias) as channel:
            return await channel.default_exchange.publish(
                aio_pika.Message(body, *args, **kwargs),
                routing_key,
                mandatory=mandatory,
                immediate=immediate,
                timeout=timeout
            )


class AioRabbitmqManager(AbsDBPoolManager):
    _clients = {}

    @staticmethod
    async def get_channel(connection_pool):
        async with connection_pool.acquire() as connection:
            return await connection.channel()

    async def create(self, alias: str, params: dict):
        if alias in self._clients:
            return self._clients[alias]

        params = params.copy()
        url = params.pop('url', None)
        assert url, "Must args url"
        connection_max_size = params.pop('connection_max_size', None)
        channel_max_size = params.pop('channel_max_size', None)

        connection_pool: Pool = Pool(aio_pika.connect_robust, url, max_size=connection_max_size)
        channel_pool: Pool = Pool(self.get_channel, connection_pool, max_size=channel_max_size)

        return self._clients.setdefault(alias, (connection_pool, channel_pool))

    def get_pool(self, alias: str):
        connection_pool, channel_pool = self._clients.get(alias)
        assert channel_pool is not None, f"Dont create the rabbitmq channel_pool named {alias}"
        assert connection_pool is not None, f"Dont create the rabbitmq connection_pool named {alias}"
        return connection_pool, channel_pool

    @asynccontextmanager
    async def get(self, alias: str):
        connection_pool, channel_pool = self.get_pool(alias)
        async with channel_pool.acquire() as channel:
            yield channel

    def executor(self, alias: str) -> RabbitmqExecutor:
        return RabbitmqExecutor(alias, self)

    async def close(self, alias: str):
        connection_pool, channel_pool = self._clients.pop(alias, (None, None))
        connection_pool: Pool
        channel_pool: Pool
        if channel_pool:
            await channel_pool.close()

        if connection_pool:
            await connection_pool.close()

    async def close_all(self):
        for alias in list(self._clients.keys()):
            await self.close(alias)

    async def from_dict(self, db_args: dict):
        for alias, rabbitmq_args in db_args.items():
            await self.create(alias, rabbitmq_args)

    async def from_settings(self, settings: aioscrapy.Settings):
        for alias, rabbitmq_args in settings.getdict('RABBITMQ_ARGS').items():
            await self.create(alias, rabbitmq_args)


rabbitmq_manager = AioRabbitmqManager()

if __name__ == '__main__':
    import asyncio


    async def test():
        _, channel_pool = await rabbitmq_manager.create('default', {
            'url': "amqp://guest:guest@192.168.234.128:5673/",
            'connection_max_size': 2,
            'channel_max_size': 10,
        })

        queue_name = 'pool_queue'
        execute = rabbitmq_manager.executor('default')

        for i in range(10):
            result = await execute.publish(queue_name, f"msg{i}".encode(), priority=3)
            print(result)

        print(await execute.get_message_count(queue_name))

        for i in range(5):
            print(await execute.get_message(queue_name))
            print(await execute.get_message_count(queue_name))

        print(await execute.clean_message_queue(queue_name))
        print(await execute.get_message_count(queue_name))

        await rabbitmq_manager.close_all()


    asyncio.get_event_loop().run_until_complete(test())



"""Download handlers for different schemes"""

from abc import abstractmethod
from typing import Optional

from aioscrapy import signals, Request, Spider
from aioscrapy.exceptions import NotConfigured, NotSupported
from aioscrapy.http import HtmlResponse
from aioscrapy.utils.httpobj import urlparse_cached
from aioscrapy.utils.log import logger
from aioscrapy.utils.misc import load_instance
from aioscrapy.utils.python import without_none_values


class BaseDownloadHandler:
    @abstractmethod
    async def download_request(self, requests: Request, spider: Spider):
        raise NotImplementedError()

    @abstractmethod
    async def close(self):
        pass


class DownloadHandlerManager:

    def __init__(self, crawler):
        self._crawler = crawler

        # stores acceptable schemes on instancing
        self._schemes: dict = without_none_values(
            crawler.settings.get('DOWNLOAD_HANDLERS_MAP', {}).get(crawler.settings.get('DOWNLOAD_HANDLERS_TYPE')) or
            crawler.settings.getwithbase('DOWNLOAD_HANDLERS')
        )
        self._handlers: dict = {}  # stores instanced handlers for schemes
        self._notconfigured: dict = {}  # remembers failed handlers
        crawler.signals.connect(self._close, signals.engine_stopped)

    @classmethod
    def from_crawler(cls, crawler) -> "DownloadHandlerManager":
        return cls(crawler)

    async def _get_handler(self, scheme: str) -> Optional[BaseDownloadHandler]:
        """Lazy-load the downloadhandler for a scheme
        only on the first request for that scheme.
        """
        if scheme in self._handlers:
            return self._handlers[scheme]
        if scheme in self._notconfigured:
            return None
        if scheme not in self._schemes:
            self._notconfigured[scheme] = 'no handler available for that scheme'
            return None

        return await self._load_handler(scheme)

    async def _load_handler(self, scheme: str) -> Optional[BaseDownloadHandler]:
        path: str = self._schemes[scheme]
        try:
            dh: BaseDownloadHandler = await load_instance(
                path,
                settings=self._crawler.settings,
            )
        except NotConfigured as ex:
            self._notconfigured[scheme] = str(ex)
            return None
        except Exception as ex:
            logger.exception(f'Loading "{path}" for scheme "{scheme}"')
            self._notconfigured[scheme] = str(ex)
            return None
        else:
            self._handlers[scheme] = dh
            return dh

    async def download_request(self, request: Request, spider: Spider) -> HtmlResponse:
        scheme = urlparse_cached(request).scheme
        handler: BaseDownloadHandler = await self._get_handler(scheme)
        if not handler:
            raise NotSupported("Unsupported URL scheme '%s': %s" %
                               (scheme, self._notconfigured[scheme]))
        return await handler.download_request(request, spider)

    async def _close(self, *_a, **_kw) -> None:
        for dh in self._handlers.values():
            await dh.close()



"""Some helpers for deprecation messages"""

import inspect
import warnings

from aioscrapy.exceptions import AioScrapyDeprecationWarning


def attribute(obj, oldattr, newattr, version='0.12'):
    cname = obj.__class__.__name__
    warnings.warn(
        f"{cname}.{oldattr} attribute is deprecated and will be no longer supported "
        f"in Aioscrapy {version}, use {cname}.{newattr} attribute instead",
        AioScrapyDeprecationWarning,
        stacklevel=3)


def create_deprecated_class(
    name,
    new_class,
    clsdict=None,
    warn_category=AioScrapyDeprecationWarning,
    warn_once=True,
    old_class_path=None,
    new_class_path=None,
    subclass_warn_message="{cls} inherits from deprecated class {old}, please inherit from {new}.",
    instance_warn_message="{cls} is deprecated, instantiate {new} instead."
):
    """
    Return a "deprecated" class that causes its subclasses to issue a warning.
    Subclasses of ``new_class`` are considered subclasses of this class.
    It also warns when the deprecated class is instantiated, but do not when
    its subclasses are instantiated.

    It can be used to rename a base class in a library. For example, if we
    have

        class OldName(SomeClass):
            # ...

    and we want to rename it to NewName, we can do the following::

        class NewName(SomeClass):
            # ...

        OldName = create_deprecated_class('OldName', NewName)

    Then, if user class inherits from OldName, warning is issued. Also, if
    some code uses ``issubclass(sub, OldName)`` or ``isinstance(sub(), OldName)``
    checks they'll still return True if sub is a subclass of NewName instead of
    OldName.
    """

    class DeprecatedClass(new_class.__class__):

        deprecated_class = None
        warned_on_subclass = False

        def __new__(metacls, name, bases, clsdict_):
            cls = super().__new__(metacls, name, bases, clsdict_)
            if metacls.deprecated_class is None:
                metacls.deprecated_class = cls
            return cls

        def __init__(cls, name, bases, clsdict_):
            meta = cls.__class__
            old = meta.deprecated_class
            if old in bases and not (warn_once and meta.warned_on_subclass):
                meta.warned_on_subclass = True
                msg = subclass_warn_message.format(cls=_clspath(cls),
                                                   old=_clspath(old, old_class_path),
                                                   new=_clspath(new_class, new_class_path))
                if warn_once:
                    msg += ' (warning only on first subclass, there may be others)'
                warnings.warn(msg, warn_category, stacklevel=2)
            super().__init__(name, bases, clsdict_)

        # see https://www.python.org/dev/peps/pep-3119/#overloading-isinstance-and-issubclass
        # and https://docs.python.org/reference/datamodel.html#customizing-instance-and-subclass-checks
        # for implementation details
        def __instancecheck__(cls, inst):
            return any(cls.__subclasscheck__(c)
                       for c in {type(inst), inst.__class__})

        def __subclasscheck__(cls, sub):
            if cls is not DeprecatedClass.deprecated_class:
                # we should do the magic only if second `issubclass` argument
                # is the deprecated class itself - subclasses of the
                # deprecated class should not use custom `__subclasscheck__`
                # method.
                return super().__subclasscheck__(sub)

            if not inspect.isclass(sub):
                raise TypeError("issubclass() arg 1 must be a class")

            mro = getattr(sub, '__mro__', ())
            return any(c in {cls, new_class} for c in mro)

        def __call__(cls, *args, **kwargs):
            old = DeprecatedClass.deprecated_class
            if cls is old:
                msg = instance_warn_message.format(cls=_clspath(cls, old_class_path),
                                                   new=_clspath(new_class, new_class_path))
                warnings.warn(msg, warn_category, stacklevel=2)
            return super().__call__(*args, **kwargs)

    deprecated_cls = DeprecatedClass(name, (new_class,), clsdict or {})

    try:
        frm = inspect.stack()[1]
        parent_module = inspect.getmodule(frm[0])
        if parent_module is not None:
            deprecated_cls.__module__ = parent_module.__name__
    except Exception as e:
        # Sometimes inspect.stack() fails (e.g. when the first import of
        # deprecated class is in jinja2 template). __module__ attribute is not
        # important enough to raise an exception as users may be unable
        # to fix inspect.stack() errors.
        warnings.warn(f"Error detecting parent module: {e!r}")

    return deprecated_cls


def _clspath(cls, forced=None):
    if forced is not None:
        return forced
    return f'{cls.__module__}.{cls.__name__}'


DEPRECATION_RULES = []


def update_classpath(path):
    """Update a deprecated path from an object with its new location"""
    for prefix, replacement in DEPRECATION_RULES:
        if isinstance(path, str) and path.startswith(prefix):
            new_path = path.replace(prefix, replacement, 1)
            warnings.warn(f"`{path}` class is deprecated, use `{new_path}` instead",
                          AioScrapyDeprecationWarning)
            return new_path
    return path



from functools import wraps

try:
    from playwright._impl._errors import Error
except ImportError:
    from playwright._impl._api_types import Error

from playwright.async_api._generated import Response as EventResponse

from aioscrapy import Request, Spider
from aioscrapy.core.downloader.handlers import BaseDownloadHandler
from aioscrapy.core.downloader.handlers.playwright.driverpool import WebDriverPool
from aioscrapy.core.downloader.handlers.playwright.webdriver import PlaywrightDriver
from aioscrapy.exceptions import DownloadError
from aioscrapy.http import PlaywrightResponse
from aioscrapy.settings import Settings
from aioscrapy.utils.tools import call_helper


class PlaywrightHandler(BaseDownloadHandler):
    def __init__(self, settings: Settings):
        self.settings = settings
        playwright_client_args = settings.getdict('PLAYWRIGHT_CLIENT_ARGS')
        use_pool = settings.getbool('PLAYWRIGHT_USE_POOL', True)
        self.wait_until = playwright_client_args.get('wait_until', 'domcontentloaded')
        self.url_regexes = playwright_client_args.pop('url_regexes', [])
        pool_size = playwright_client_args.pop('pool_size', settings.getint("CONCURRENT_REQUESTS", 1))
        self._webdriver_pool = WebDriverPool(use_pool=use_pool, pool_size=pool_size, driver_cls=PlaywrightDriver, **playwright_client_args)

    @classmethod
    def from_settings(cls, settings: Settings):
        return cls(settings)

    async def download_request(self, request: Request, spider: Spider) -> PlaywrightResponse:
        try:
            return await self._download_request(request, spider)
        except Error as e:
            raise DownloadError(real_error=e) from e

    async def _download_request(self, request: Request, spider) -> PlaywrightResponse:
        cookies = dict(request.cookies)
        timeout = request.meta.get('download_timeout', 30) * 1000
        user_agent = request.headers.get("User-Agent")
        proxy: str = request.meta.get("proxy")
        url = request.url

        cache_response = {}

        # 为了获取监听事件中的响应结果
        def on_event_wrap_handler(func):
            @wraps(func)
            async def inner(response):
                ret = await func(response)
                if ret:
                    cache_response[ret[0]] = ret[1]

            return inner

        kwargs = dict()
        if proxy:
            kwargs['proxy'] = proxy
        if user_agent:
            kwargs['user_agent'] = user_agent

        driver: PlaywrightDriver = await self._webdriver_pool.get(**kwargs)

        # 移除所有的事件监听事件后 重新添加
        driver.page._events = dict()
        for name in dir(spider):
            if not name.startswith('on_event_'):
                continue
            driver.page.on(name.replace('on_event_', ''), on_event_wrap_handler(getattr(spider, name)))

        try:
            if cookies:
                driver.url = url
                await driver.set_cookies(cookies)
            await driver.page.goto(url, wait_until=request.meta.get('wait_until', self.wait_until), timeout=timeout)

            if process_action_fn := getattr(spider, 'process_action', None):
                action_result = await call_helper(process_action_fn, driver)
                if action_result:
                    cache_response[action_result[0]] = action_result[1]

            for cache_key in list(cache_response.keys()):
                if isinstance(cache_response[cache_key], EventResponse):
                    cache_ret = cache_response[cache_key]
                    cache_response[cache_key] = PlaywrightResponse(
                        url=cache_ret.url,
                        request=request,
                        intercept_request=dict(
                            url=cache_ret.request.url,
                            headers=cache_ret.request.headers,
                            data=cache_ret.request.post_data,
                        ),
                        headers=cache_ret.headers,
                        body=await cache_ret.body(),
                        status=cache_ret.status,
                    )

            return PlaywrightResponse(
                url=driver.page.url,
                status=200,
                text=await driver.page.content(),
                cookies=await driver.get_cookies(),
                cache_response=cache_response,
                driver=driver,
                driver_pool=self._webdriver_pool
            )
        except Exception as e:
            await self._webdriver_pool.remove(driver)
            raise e

    async def close(self):
        await self._webdriver_pool.close()



from curl_cffi.curl import CurlError
from curl_cffi.requests import AsyncSession

from aioscrapy import Request
from aioscrapy.core.downloader.handlers import BaseDownloadHandler
from aioscrapy.exceptions import DownloadError
from aioscrapy.http import HtmlResponse
from aioscrapy.settings import Settings
from aioscrapy.utils.log import logger


class CurlCffiDownloadHandler(BaseDownloadHandler):

    def __init__(self, settings):
        self.settings: Settings = settings
        self.httpx_client_session_args: dict = self.settings.get('CURL_CFFI_CLIENT_SESSION_ARGS', {})
        self.verify_ssl: bool = self.settings.get("VERIFY_SSL", True)

    @classmethod
    def from_settings(cls, settings: Settings):
        return cls(settings)

    async def download_request(self, request: Request, _) -> HtmlResponse:
        try:
            return await self._download_request(request)
        except CurlError as e:
            raise DownloadError(real_error=e) from e

    async def _download_request(self, request: Request) -> HtmlResponse:
        kwargs = {
            'timeout': self.settings.get('DOWNLOAD_TIMEOUT'),
            'cookies': dict(request.cookies),
            'verify': request.meta.get('verify_ssl', self.verify_ssl),
            'allow_redirects': self.settings.getbool('REDIRECT_ENABLED', True) if request.meta.get(
                'dont_redirect') is None else request.meta.get('dont_redirect'),
            'impersonate': request.meta.get('impersonate'),
        }
        post_data = request.body or None
        if isinstance(post_data, dict):
            kwargs['json'] = post_data
        else:
            kwargs['data'] = post_data

        headers = request.headers or self.settings.get('DEFAULT_REQUEST_HEADERS')
        kwargs['headers'] = headers

        proxy = request.meta.get("proxy")
        if proxy:
            kwargs["proxies"] = {'http': proxy, 'https': proxy}
            logger.debug(f"use proxy {proxy}: {request.url}")

        session_args = self.httpx_client_session_args.copy()

        async with AsyncSession(**session_args) as session:
            response = await session.request(request.method, request.url, **kwargs)

        return HtmlResponse(
            str(response.url),
            status=response.status_code,
            headers=response.headers,
            body=response.content,
            cookies={j.name: j.value or '' for j in response.cookies.jar},
            encoding=response.encoding
        )

    async def close(self):
        pass



import asyncio
import math
from io import BytesIO
from typing import Tuple, Optional

import requests
import xlsxwriter
from PIL import Image, ImageFile

from aioscrapy.utils.log import logger

try:
    resample = Image.LANCZOS
except:
    resample = Image.ANTIALIAS
ImageFile.LOAD_TRUNCATED_IMAGES = True


class ExeclSinkMixin:
    ws_cache = {}
    wb_cache = {}
    fields_cache = {}
    y_cache = {}

    @staticmethod
    async def deal_img(url: str, img_size: Optional[Tuple[int, int]]) -> Optional[BytesIO]:
        if url.startswith('//'):
            url = 'https:' + url
        try:
            img_bytes = requests.get(url).content
        except Exception as e:
            logger.error(f"download img error: {e}")
            return
        im = Image.open(BytesIO(img_bytes))
        im_format = im.format
        if img_size:
            temp = max(im.size[0] / img_size[0], im.size[1] / img_size[1])
            img_size = (math.ceil(im.size[0] / temp), math.ceil(im.size[1] / temp))
            im = im.resize(img_size, resample).convert('P')
        result = BytesIO()
        im.save(result, format=im_format)
        return result

    async def save_item(
            self,
            item: dict,
            *,
            filename: Optional[str] = None,
            date_fields: Optional[list] = None,
            date_format: str = 'yyyy-mm-dd HH:MM:SS',
            img_fields: Optional[list] = None,
            img_size: Optional[Tuple[int, int]] = None,
            **options
    ):
        assert filename is not None, "请传入filename参数"
        if '.xlsx' not in filename:
            filename = filename + '.xlsx'
        try:
            wb, ws, fields, y = self._get_write_class(filename, item, **options)
            bold_format_1 = wb.add_format({'align': 'left', 'border': 1, 'valign': 'vcenter'})
            bold_format_2 = wb.add_format({'align': 'left', 'border': 1, 'valign': 'vcenter', 'fg_color': '#D0D3D4'})
            for x, field in enumerate(fields):
                if x % 2 == 0:
                    bold_format = bold_format_1
                else:
                    bold_format = bold_format_2
                if date_fields is not None and field in date_fields:
                    ws.write_datetime(y, x, item.get(field), wb.add_format({'num_format': date_format}))

                elif img_fields is not None and field in img_fields:
                    img_size and ws.set_column_pixels(x, x, width=math.ceil(img_size[0]))
                    url = item.get(field)
                    img_bytes = await self.deal_img(url, img_size)
                    if img_bytes is None or ws.insert_image(y, x, '', {'image_data': img_bytes}) == -1:
                        ws.write(y, x, url, bold_format)
                else:
                    ws.write(y, x, item.get(field), bold_format)
            if img_size is not None:
                ws.set_column_pixels(0, len(fields), width=math.ceil(img_size[0]))
                ws.set_row_pixels(y, height=math.ceil(img_size[1]))
        except Exception as e:
            logger.exception(f'Save Execl Error, filename:{filename}, item:{item}, errMsg: {e}')

    def _get_write_class(self, filename, item, sheet='sheet1', **options):
        filename_sheet = filename + sheet
        if self.ws_cache.get(filename_sheet) is None:
            if self.wb_cache.get(filename) is None:
                logger.info(f'Create Execl: {filename}')
                wb = xlsxwriter.Workbook(filename, options=options)
                self.wb_cache[filename] = wb
            else:
                wb = self.wb_cache[filename]
            ws = wb.add_worksheet(sheet)
            bold_format = wb.add_format(
                {'bold': True, 'font_size': 12, 'border': 1, 'align': 'center', 'valign': 'vcenter'})
            fields = list(item.keys())
            ws.write_row('A1', fields, cell_format=bold_format)
            ws.set_row(0, height=30)
            self.fields_cache[filename_sheet] = fields
            self.ws_cache[filename_sheet] = ws
            self.y_cache[filename_sheet] = 0
        self.y_cache[filename_sheet] += 1
        return self.wb_cache[filename], \
            self.ws_cache[filename_sheet], \
            self.fields_cache[filename_sheet], \
            self.y_cache[filename_sheet]

    def close_execl(self, filename=None):
        if filename not in self.wb_cache:
            return

        logger.info(f'Closing Execl: {filename}')
        if wb := self.wb_cache.pop(filename):
            wb.close()
        for filename_sheet in list(self.ws_cache.keys()):
            if not filename_sheet.startswith(filename):
                continue
            self.ws_cache.pop(filename_sheet, None)
            self.y_cache.pop(filename_sheet, None)
            self.fields_cache.pop(filename_sheet, None)

    def close(self):
        for filename in list(self.wb_cache.keys()):
            self.close_execl(filename)


class ExeclPipeline(ExeclSinkMixin):
    def __init__(self, settings):
        self.lock = asyncio.Lock()

    @classmethod
    def from_settings(cls, settings):
        return cls(settings)

    async def process_item(self, item, spider):
        execl_kw: Optional[dict] = item.pop('__execl__', None)
        if not execl_kw:
            logger.warning(f"item Missing key __execl__, not stored")
            return item

        execl_kw.setdefault('filename', spider.name)
        async with self.lock:
            await self.save_item(item, **execl_kw)

    async def close_spider(self, spider):
        self.close()


if __name__ == '__main__':
    class TestSpider:
        name = 'TestSpider'


    async def test():
        p = ExeclPipeline({})
        await p.process_item({
            'title': 'tttt',
            'img': '//www.baidu.com/img/flexible/logo/pc/result.png',
            '__execl__': {
                'sheet': 'sheet1',
                # 'filename': 'test',
                # 'img_fields': ['img'],
                # 'img_size': (100, 500)
            }
        }, TestSpider())
        await p.close_spider(None)


    asyncio.run(test())



from abc import ABC
from typing import Optional

import aioscrapy
from aioscrapy.db import db_manager
from aioscrapy.queue import AbsQueue
from aioscrapy.serializer import AbsSerializer
from aioscrapy.utils.misc import load_object


class RedisQueueBase(AbsQueue, ABC):
    inc_key = 'scheduler/enqueued/redis'

    @classmethod
    def from_dict(cls, data: dict) -> "RedisQueueBase":
        alias: str = data.get("alias", 'queue')
        server: aioscrapy.db.aioredis.Redis = db_manager.redis(alias)
        spider_name: str = data["spider_name"]
        serializer: str = data.get("serializer", "aioscrapy.serializer.JsonSerializer")
        serializer: AbsSerializer = load_object(serializer)
        return cls(
            server,
            key='%(spider)s:requests' % {'spider': spider_name},
            serializer=serializer
        )

    @classmethod
    async def from_spider(cls, spider: aioscrapy.Spider) -> "RedisQueueBase":
        alias: str = spider.settings.get("SCHEDULER_QUEUE_ALIAS", 'queue')
        server: aioscrapy.db.aioredis.Redis = db_manager.redis(alias)
        queue_key: str = spider.settings.get("SCHEDULER_QUEUE_KEY", '%(spider)s:requests')
        serializer: str = spider.settings.get("SCHEDULER_SERIALIZER", "aioscrapy.serializer.JsonSerializer")
        serializer: AbsSerializer = load_object(serializer)
        return cls(
            server,
            spider,
            queue_key % {'spider': spider.name},
            serializer=serializer
        )

    async def clear(self) -> None:
        """Clear queue/stack"""
        await self.container.delete(self.key)


class RedisFifoQueue(RedisQueueBase):
    """Per-spider FIFO queue"""

    async def len(self) -> int:
        return await self.container.llen(self.key)

    async def push(self, request: aioscrapy.Request) -> None:
        """Push a request"""
        await self.container.lpush(self.key, self._encode_request(request))

    async def push_batch(self, requests) -> None:
        async with self.container.pipeline() as pipe:
            for request in requests:
                pipe.lpush(self.key, self._encode_request(request))
            await pipe.execute()

    async def pop(self, count: int = 1) -> Optional[aioscrapy.Request]:
        """Pop a request"""
        async with self.container.pipeline(transaction=True) as pipe:
            for _ in range(count):
                pipe.rpop(self.key)
            results = await pipe.execute()
        for result in results:
            if result:
                yield await self._decode_request(result)


class RedisPriorityQueue(RedisQueueBase):
    """Per-spider priority queue abstraction using redis' sorted set"""

    async def len(self) -> int:
        return await self.container.zcard(self.key)

    async def push(self, request: aioscrapy.Request) -> None:
        """Push a request"""
        data = self._encode_request(request)
        score = request.priority
        await self.container.zadd(self.key, {data: score})

    async def push_batch(self, requests) -> None:
        async with self.container.pipeline() as pipe:
            for request in requests:
                pipe.zadd(self.key, {self._encode_request(request): request.priority})
            await pipe.execute()

    async def pop(self, count: int = 1) -> Optional[aioscrapy.Request]:
        async with self.container.pipeline(transaction=True) as pipe:
            stop = count - 1 if count - 1 > 0 else 0
            results, _ = await (
                pipe.zrange(self.key, 0, stop)
                .zremrangebyrank(self.key, 0, stop)
                .execute()
            )
        for result in results:
            yield await self._decode_request(result)


class RedisLifoQueue(RedisQueueBase):
    """Per-spider LIFO queue."""

    async def len(self) -> int:
        return await self.container.llen(self.key)

    async def push(self, request: aioscrapy.Request) -> None:
        """Push a request"""
        await self.container.lpush(self.key, self._encode_request(request))

    async def push_batch(self, requests) -> None:
        async with self.container.pipeline() as pipe:
            for request in requests:
                pipe.lpush(self.key, self._encode_request(request))
            await pipe.execute()

    async def pop(self, count: int = 1) -> Optional[aioscrapy.Request]:
        """Pop a request"""
        async with self.container.pipeline(transaction=True) as pipe:
            for _ in range(count):
                pipe.lpop(self.key)
            results = await pipe.execute()
        for result in results:
            if result:
                yield await self._decode_request(result)


SpiderQueue = RedisFifoQueue
SpiderStack = RedisLifoQueue
SpiderPriorityQueue = RedisPriorityQueue



"""
This module contains essential stuff that should've come with Python itself ;)
"""
import gc
import re
import sys
import weakref
from functools import wraps

from aioscrapy.utils.decorators import deprecated


def is_listlike(x):
    """
    >>> is_listlike("foo")
    False
    >>> is_listlike(5)
    False
    >>> is_listlike(b"foo")
    False
    >>> is_listlike([b"foo"])
    True
    >>> is_listlike((b"foo",))
    True
    >>> is_listlike({})
    True
    >>> is_listlike(set())
    True
    >>> is_listlike((x for x in range(3)))
    True
    >>> is_listlike(range(5))
    True
    """
    return hasattr(x, "__iter__") and not isinstance(x, (str, bytes))


def to_unicode(text, encoding=None, errors='strict'):
    """Return the unicode representation of a bytes object ``text``. If
    ``text`` is already an unicode object, return it as-is."""
    if isinstance(text, str):
        return text
    if not isinstance(text, (bytes, str)):
        raise TypeError('to_unicode must receive a bytes or str '
                        f'object, got {type(text).__name__}')
    if encoding is None:
        encoding = 'utf-8'
    return text.decode(encoding, errors)


def to_bytes(text, encoding=None, errors='strict'):
    """Return the binary representation of ``text``. If ``text``
    is already a bytes object, return it as-is."""
    if isinstance(text, bytes):
        return text
    if not isinstance(text, str):
        raise TypeError('to_bytes must receive a str or bytes '
                        f'object, got {type(text).__name__}')
    if encoding is None:
        encoding = 'utf-8'
    return text.encode(encoding, errors)


@deprecated('to_unicode')
def to_native_str(text, encoding=None, errors='strict'):
    """ Return str representation of ``text``. """
    return to_unicode(text, encoding, errors)


def re_rsearch(pattern, text, chunk_size=1024):
    """
    This function does a reverse search in a text using a regular expression
    given in the attribute 'pattern'.
    Since the re module does not provide this functionality, we have to find for
    the expression into chunks of text extracted from the end (for the sake of efficiency).
    At first, a chunk of 'chunk_size' kilobytes is extracted from the end, and searched for
    the pattern. If the pattern is not found, another chunk is extracted, and another
    search is performed.
    This process continues until a match is found, or until the whole file is read.
    In case the pattern wasn't found, None is returned, otherwise it returns a tuple containing
    the start position of the match, and the ending (regarding the entire text).
    """

    def _chunk_iter():
        offset = len(text)
        while True:
            offset -= (chunk_size * 1024)
            if offset <= 0:
                break
            yield (text[offset:], offset)
        yield (text, 0)

    if isinstance(pattern, str):
        pattern = re.compile(pattern)

    for chunk, offset in _chunk_iter():
        matches = [match for match in pattern.finditer(chunk)]
        if matches:
            start, end = matches[-1].span()
            return offset + start, offset + end
    return None


def memoizemethod_noargs(method):
    """Decorator to cache the result of a method (without arguments) using a
    weak reference to its object
    """
    cache = weakref.WeakKeyDictionary()

    @wraps(method)
    def new_method(self, *args, **kwargs):
        if self not in cache:
            cache[self] = method(self, *args, **kwargs)
        return cache[self]

    return new_method


def without_none_values(iterable):
    """Return a copy of ``iterable`` with all ``None`` entries removed.

    If ``iterable`` is a mapping, return a dictionary where all pairs that have
    value ``None`` have been removed.
    """
    try:
        return {k: v for k, v in iterable.items() if v is not None}
    except AttributeError:
        return type(iterable)((v for v in iterable if v is not None))


def global_object_name(obj):
    """
    Return full name of a global object.

    >>> from aioscrapy import Request
    >>> global_object_name(Request)
    'aioscrapy.http.request.Request'
    """
    return f"{obj.__module__}.{obj.__name__}"


if hasattr(sys, "pypy_version_info"):
    def garbage_collect():
        # Collecting weakreferences can take two collections on PyPy.
        gc.collect()
        gc.collect()
else:
    def garbage_collect():
        gc.collect()



import asyncio

from aioscrapy.utils.log import logger
from aioscrapy.utils.tools import create_task


class SqlFormat:

    @staticmethod
    def pg_insert(table: str, fields: list, *args) -> str:
        placeholder = ','.join([f'${i + 1}' for i in range(len(fields))])
        return f'''INSERT INTO {table} ({",".join(fields)}) VALUES ({placeholder})'''

    @staticmethod
    def pg_ignore_insert(table: str, fields: list, *args) -> str:
        placeholder = ','.join([f'${i + 1}' for i in range(len(fields))])
        return f'''INSERT INTO {table} ({",".join(fields)}) VALUES ({placeholder}) ON CONFLICT DO NOTHING'''

    @staticmethod
    def pg_update_insert(table: str, fields: list, update_fields: list, on_conflict: str, *args) -> str:
        assert on_conflict is not None, "on_conflict must be str, eg: 'id'"
        placeholder = ','.join([f'${i + 1}' for i in range(len(fields))])
        if not update_fields:
            update_fields = fields
        update_fields = ','.join([f"{key} = excluded.{key}" for key in update_fields])
        return f'''INSERT INTO {table} ({",".join(fields)}) VALUES ({placeholder}) ON CONFLICT({on_conflict}) DO UPDATE SET {update_fields}'''

    @staticmethod
    def mysql_insert(table: str, fields: list, *args) -> str:
        placeholder = ','.join(['%s'] * len(fields))
        fields = ','.join(fields)
        return f'''INSERT INTO {table} ({fields}) VALUES ({placeholder})'''

    @staticmethod
    def mysql_ignore_insert(table: str, fields: list, *args) -> str:
        placeholder = ','.join(['%s'] * len(fields))
        fields = ','.join(fields)
        return f'''INSERT IGNORE INTO {table} ({fields}) VALUES ({placeholder})'''

    @staticmethod
    def mysql_update_insert(table: str, fields: list, update_fields: list, *args) -> str:
        placeholder = ','.join(['%s'] * len(fields))
        if not update_fields:
            update_fields = fields
        update_fields = ','.join([f"{key} = VALUES({key})" for key in update_fields])
        fields = ','.join(fields)
        return f'INSERT INTO {table} ({fields}) VALUES ({placeholder}) ON DUPLICATE KEY UPDATE {update_fields}'

    def __call__(self, *args, db_type='mysql', insert_type='insert'):
        if getattr(self, f'{db_type}_{insert_type}'):
            return getattr(self, f'{db_type}_{insert_type}')(*args)
        raise Exception(f"This write type is not supported： {db_type}_{insert_type}")


get_sql = SqlFormat()


class ItemCacheMixin:
    def __init__(self, db_type: str):
        self.db_type = db_type
        self.item_cache = {}
        self.fields_cache = {}
        self.table_cache = {}
        self.insert_sql_cache = {}
        self.db_alias_cache = {}

    def parse_item_to_cache(self, item: dict, save_info):
        table_name = save_info.get('table_name')
        assert table_name is not None, 'Missing table_name'
        insert_type = save_info.get('insert_type', 'insert')
        update_fields = save_info.get('update_fields', [])
        db_alias = save_info.get('db_alias', ['default'])
        on_conflict = save_info.get('on_conflict')
        if isinstance(db_alias, str):
            db_alias = [db_alias]

        fields = list(item.keys())
        cache_key = ''.join(fields + update_fields + db_alias) + insert_type + table_name + (on_conflict or '')

        if self.fields_cache.get(cache_key) is None:
            self.db_alias_cache[cache_key] = db_alias
            self.table_cache[cache_key] = table_name
            self.fields_cache[cache_key] = fields
            self.item_cache[cache_key] = []
            self.insert_sql_cache[cache_key] = get_sql(
                table_name, fields, update_fields, on_conflict,
                db_type=self.db_type,
                insert_type=insert_type,
            )

        self.item_cache[cache_key].append([item[field] for field in self.fields_cache[cache_key]])
        return cache_key, len(self.item_cache[cache_key])


class DBPipelineBase(ItemCacheMixin):
    def __init__(self, settings, db_type: str):
        super().__init__(db_type)
        self.cache_num = settings.getint('SAVE_CACHE_NUM', 500)
        self.save_cache_interval = settings.getint('SAVE_CACHE_INTERVAL', 10)
        self.lock = asyncio.Lock()
        self.running: bool = True
        self.item_save_key: str = f'__{db_type}__'

    async def open_spider(self, spider):
        create_task(self.save_heartbeat())

    async def save_heartbeat(self):
        while self.running:
            await asyncio.sleep(self.save_cache_interval)
            create_task(self.save_all())

    async def process_item(self, item, spider):
        save_info = item.pop(self.item_save_key, None)
        if save_info is None:
            logger.warning(f"item Missing key {self.item_save_key}, not stored")
            return item

        await self.save_item(item, save_info)
        return item

    async def close_spider(self, spider):
        self.running = False
        await self.save_all()

    async def save_all(self):
        async with self.lock:
            for cache_key, items in self.item_cache.items():
                items and await self._save(cache_key)

    async def save_item(self, item: dict, save_info: dict):
        async with self.lock:
            cache_key, cache_count = self.parse_item_to_cache(item, save_info)
            if cache_count >= self.cache_num:
                await self._save(cache_key)

    async def _save(self, cache_key):
        raise NotImplementedError



"""
Url Length Spider Middleware

See documentation in docs/topics/spider-middleware.rst
"""

from aioscrapy.exceptions import NotConfigured
from aioscrapy.http import Request
from aioscrapy.utils.log import logger


class UrlLengthMiddleware:

    def __init__(self, maxlength):
        self.maxlength = maxlength

    @classmethod
    def from_settings(cls, settings):
        maxlength = settings.getint('URLLENGTH_LIMIT')
        if not maxlength:
            raise NotConfigured
        return cls(maxlength)

    async def process_spider_output(self, response, result, spider):
        def _filter(request):
            if isinstance(request, Request) and len(request.url) > self.maxlength:
                logger.info(
                    "Ignoring link (url length > %(maxlength)d): %(url)s " % {
                        'maxlength': self.maxlength, 'url': request.url
                    }
                )
                spider.crawler.stats.inc_value('urllength/request_ignored_count', spider=spider)
                return False
            else:
                return True

        return (r async for r in result or () if _filter(r))
